MODULE PZCompCands EXPORTS Main;

(* Compares files of "found" and "ideal" candidates, writes diff and inter. *)

(* Last edited on 1999-11-18 03:32:43 by hcgl *)

IMPORT ParseParams, Text, FileRd, FileWr, Wr, Fmt;
IMPORT Process, OSError, Thread, Stdio;
IMPORT PZCandidate, PZMatch;

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, INT, NAT, BOOL, BOOLS;

<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  Options = RECORD
      found: TEXT;           (* Input file name of "found" candidates. *)
      ideal: TEXT;           (* Input file name of "ideal" candidates. *)
      good: TEXT;            (* Output file name of "good" candidates. *)
      missing: TEXT;         (* Output file name of "missing" candidates. *)
      bogus: TEXT;           (* Output file name of "bogus" candidates. *)
      step: LONG;            (* Nominal sampling step. *)

      printMatches: BOOL;    (* TRUE prints overlapping (ideal,found) pairs. *)
      printMissing: BOOL;    (* TRUE prints ideal cands that aren't found cands. *)
      printBogus: BOOL;      (* TRUE prints found cands that aren't ideal cands. *)
      printCurvePairs: BOOL; (* TRUE prints summary line per ideal curve pair. *)
      
      (* Criteria for ideal candidate selection: *)
      minIdealLength: LONG;  (* Min length (pixels) of ideal cands. *)
      blurFactor: LONG;      (* Corner broadening factor. *)
      
      (* Criteria for equality of candidates: *) 
      minCompOverlap: LONG;  (* Minimum overlap (pixels). *)
      maxCompShift: LONG;    (* Maximum displacement on each chain. *)
    END;

PROCEDURE Main() =
  VAR iNOK, fNOK: NAT;
  BEGIN
    WITH
      o = GetOptions(),

      fcData = ReadFoundCandidates(o.found),
      fCand = fcData.c^,
      lambda = fcData.lambda,
      
      minIdealLength = MAX(0.0d0, o.minIdealLength - 2.0d0*o.blurFactor*lambda),
      minIdealSteps = FLOOR(minIdealLength/o.step),
      icData = ReadIdealCandidates(o.ideal, minIdealSteps),
      iCand = icData.c^
    DO
      Wr.PutText(stderr, 
        "ignoring ideal cands that cover less than " & 
        FI(minIdealSteps,1) & " sampling steps\n"
      );
      <* ASSERT icData.lambda = fcData.lambda *>
      ProcessCandidates(
        iCand, fCand,
        minSteps := CEILING(o.minCompOverlap/o.step - 0.0001d0),
        maxAdjust := CEILING(o.maxCompShift/o.step - 0.0001d0),
        printMatches := o.printMatches, 
        printMissing := o.printMissing, 
        printBogus := o.printBogus,
        printCurvePairs := o.printCurvePairs,
        (*OUT*) iNOK := iNOK, fNOK := fNOK
      );
      
      WriteCandidateSubset(
        o.good, SUBARRAY(fCand, 0, fNOK),
        "found candidates that match some ideal candidate",
        fcData.cmt, fcData.lambda
      );
      
      WriteCandidateSubset(
        o.bogus, SUBARRAY(fCand, fNOK, NUMBER(fCand) - fNOK),
        "found candidates that don't match any ideal candidates",
        fcData.cmt, fcData.lambda
      );
      
      WriteCandidateSubset(
        o.missing, SUBARRAY(iCand, iNOK, NUMBER(iCand) - iNOK),
        "ideal candidates that don't match any found candidates",
        icData.cmt, icData.lambda
      );
    END;
  END Main;

PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-found");
        o.found := pp.getNext();

        pp.getKeyword("-ideal");
        o.ideal := pp.getNext();

        IF pp.keywordPresent("-good") THEN
          o.good := pp.getNext()
        ELSE
          o.good := ""
        END;

        IF pp.keywordPresent("-missing") THEN
          o.missing := pp.getNext()
        ELSE
          o.missing := ""
        END;

        IF pp.keywordPresent("-bogus") THEN
          o.bogus := pp.getNext()
        ELSE
          o.bogus := ""
        END;

        pp.getKeyword("-step");
        o.step := pp.getNextLongReal(0.0d0, 1024.0d0);
        
        IF pp.keywordPresent("-minIdealLength") THEN
          o.minIdealLength := pp.getNextLongReal(0.0d0)
        ELSE
          o.minIdealLength := 0.0d0
        END;

        pp.getKeyword("-blurFactor");
        o.blurFactor := pp.getNextLongReal(0.01d0, 100.0d0);

        pp.getKeyword("-minCompOverlap");
        o.minCompOverlap := pp.getNextLongReal(0.0d0);

        pp.getKeyword("-maxCompShift");
        o.maxCompShift := pp.getNextLongReal(0.0d0);

        o.printMatches    := pp.keywordPresent("-printMatches");
        o.printMissing    := pp.keywordPresent("-printMissing");
        o.printBogus      := pp.keywordPresent("-printBogus");
        o.printCurvePairs := pp.keywordPresent("-printCurvePairs");

        pp.finish();
      EXCEPT
      | ParseParams.Error =>
          Wr.PutText(stderr, "Usage: PZCompCands \\\n");
          Wr.PutText(stderr, "  -found NAME \\\n");
          Wr.PutText(stderr, "  -ideal NAME \\\n");
          Wr.PutText(stderr, "  -good NAME \\\n");
          Wr.PutText(stderr, "  -missing NAME \\\n");
          Wr.PutText(stderr, "  -step PIXELS \\\n");
          Wr.PutText(stderr, "  [ -minIdealLength PIXELS ] \\\n");
          Wr.PutText(stderr, "  -blurFactor PIXELS \\\n");
          Wr.PutText(stderr, "  -minCompOverlap PIXELS -maxCompShift PIXELS \\\n");
          Wr.PutText(stderr, "  [ -printMatches ] \\\n");
          Wr.PutText(stderr, "  [ -printMissing ] \\\n");
          Wr.PutText(stderr, "  [ -printBogus ] \\\n");
          Wr.PutText(stderr, "  [ -printCurvePairs ]\n");
      Process.Exit(1);
      END;
    END;
    RETURN o
  END GetOptions;
  
PROCEDURE ReadFoundCandidates(name: TEXT): PZCandidate.ReadData =
  (* Reads a candidate list. *) 
  BEGIN
    RETURN PZCandidate.Read(FileRd.Open(name & ".can"))
  END ReadFoundCandidates;
  
PROCEDURE ReadIdealCandidates(name: TEXT; minSteps: NAT): PZCandidate.ReadData =
  (* 
    Reads a candidate list, and removes entries with less than 
    "minSteps" sampling steps on either segment. *) 
  VAR nCands: NAT;
  BEGIN
    WITH
      cData = PZCandidate.Read(FileRd.Open(name & ".can")),
      c = cData.c^
    DO
      nCands := NUMBER(c);
      PZCandidate.RemoveShort(c, nCands, minSteps);
      WITH 
        rd = NEW(REF PZCandidate.List, nCands), d = rd^,
        cmt = cData.cmt & "\n" &
          "PZCompCands: \n" &
          "  removed candidates with less than " & FI(minSteps,5) & " steps\n"
      DO
        d := SUBARRAY(c, 0, nCands);
        RETURN PZCandidate.ReadData{cmt := cmt, lambda := cData.lambda, c := rd}
      END;
    END;
  END ReadIdealCandidates;

PROCEDURE ProcessCandidates(
    VAR (*IO*) iCand: PZCandidate.List;
    VAR (*IO*) fCand: PZCandidate.List;
    minSteps: NAT;
    maxAdjust: NAT;
    printMatches: BOOL;
    printMissing: BOOL;
    printBogus: BOOL;
    printCurvePairs: BOOL;
    VAR (*OUT*) iNOK: NAT;
    VAR (*OUT*) fNOK: NAT;
  )=
  (*
    Compares the `found' candidates "fCand" against the `ideal'
    candidates "iCand".  An found candidate that overlaps an ideal one
    (with approximately the same alignment, up to "tolerance" samples)
    is said to be `good', otherwise it is `bogus'.  An ideal candidate
    that overlaps an found candidate is said to be `present', otherwise
    it is `missing'.  Returns in "iNOK" the number of ideal cands that
    are present, and in "fNOK" the number of found cands that are good.
    Each found list is re-sorted, so that the good cands are
    "fCand[0..fNOK-1]" and the present candidates are "iCand[0..iNOK-1]".
  *)

  VAR
    iIni, fIni: NAT;
    iFin, fFin: NAT;

  PROCEDURE ComparePairs(READONLY a, b: PZCandidate.T): [-1..+1] =
    (*
      Compares the curve pair of "a" with the curve pair of "b", 
      lexicographically.  Returns 0 iff the two candidates belong
      to the same pair of curves.
    *)
    BEGIN
      WITH
        k0a = a.seg[0].cvx,
        k1a = a.seg[1].cvx,
        k0b = b.seg[0].cvx,
        k1b = b.seg[1].cvx
      DO
        IF k0a < k0b THEN
          RETURN -1
        ELSIF k0a > k0b THEN
          RETURN +1
        ELSIF k1a < k1b THEN
          RETURN -1
        ELSIF k1a > k1b THEN
          RETURN +1
        ELSE
          RETURN 0
        END
      END
    END ComparePairs;

  PROCEDURE SortOK(VAR c: PZCandidate.List; VAR OK: BOOLS) =
    (*
      Rearranges the list "c" and the bit vector "OK" 
      simultaneously so as to bring all the elements
      with "OK = TRUE" to the beginning. 
    *)
    VAR r, s: NAT;
    BEGIN
      r := 0; s := LAST(c);
      WHILE r < s DO
        (* Invariant: OK[0..r-1] = TRUE, OK[s+1..LAST(c)] = FALSE *)
        (* Look for swappable pair: *)
        WHILE r < s AND OK[r] DO INC(r) END;
        WHILE r < s AND NOT OK[s] DO DEC(s) END;
        IF r < s THEN
          VAR t := c[r]; BEGIN c[r] := c[s]; c[s] := t END;
          VAR t := OK[r]; BEGIN OK[r] := OK[s]; OK[s] := t END;
          INC(r); DEC(s);
        END
      END
    END SortOK;
    
  VAR ra, rb: REF PZMatch.T := NIL;  (* Work areas for "PZCandidate.Overlap" *)

  BEGIN
    WITH
      iN = NUMBER(iCand),
      fN = NUMBER(fCand),
      iOK = NEW(REF BOOLS, iN)^,
      fOK = NEW(REF BOOLS, fN)^
    DO
      PZCandidate.Sort(iCand, PZCandidate.LexicallyBetter);
      PZCandidate.Sort(fCand, PZCandidate.LexicallyBetter);
      FOR fIndex := 0 TO fN-1 DO fOK[fIndex] := FALSE END;
      FOR iIndex := 0 TO iN-1 DO iOK[iIndex] := FALSE END;
      iIni := 0; fIni := 0;
      WHILE iIni < iN DO
        (* Another curve pair in the ideal set... *)

        (* Paranoid check of sorting: *)
        IF iIni > 0 AND iIni < iN THEN 
          <* ASSERT PZCandidate.LexicallyBetter(iCand[iIni-1], iCand[iIni]) *>
        END;

        (* Locate all ideal candidates "[iIni..iFin-1]" with this curve pair: *)
        iFin := iIni+1;
        WHILE iFin < iN
        AND ComparePairs(iCand[iIni], iCand[iFin]) = 0
        DO
          INC(iFin)
        END;
        
        (* Locate all found candidates "[fIni..fFin-1]" with this curve pair: *)
        WHILE fIni < fN
        AND ComparePairs(fCand[fIni], iCand[iIni]) < 0
        DO
          INC(fIni)
        END;
        fFin := fIni;
        WHILE fFin < fN
        AND ComparePairs(iCand[iIni], fCand[fFin]) = 0
        DO
          INC(fFin)
        END;
        
        IF printCurvePairs THEN
          WITH s = iCand[iIni].seg DO
            Wr.PutText(stderr, "curves: " & FI(s[0].cvx,4) & "  " & FI(s[1].cvx,4));
          END;
          Wr.PutText(stderr, ": " & FI(iFin - iIni,6) & " ideal cands, ");
          Wr.PutText(stderr, FI(fFin - fIni,6) & " found cands\n");
        END;
        
        (* Compare each ideal Cand against all found Cands with this curve pair: *)
        IF fIni < fFin AND fIni < fN THEN
          ProcessCandidatesOfCurvePair(
            iIni,
            SUBARRAY(iCand, iIni, iFin-iIni),
            SUBARRAY(iOK, iIni, iFin-iIni),
            fIni,
            SUBARRAY(fCand, fIni, fFin-fIni),
            SUBARRAY(fOK, fIni, fFin-fIni),
            minSteps, maxAdjust,
            printMatches,
            ra, rb
          );
        END;

        (* Start next curve pair: *)
        iIni := iFin; fIni := fFin
      END;

      (* Count matched ideal cands and print missing ones: *)
      IF printMissing THEN Wr.PutText(stderr, "=== ideal NOT in found ===\n") END;
      iNOK := 0;
      FOR iIndex := 0 TO iN-1 DO
        IF iOK[iIndex] THEN
          INC(iNOK)
        ELSIF printMissing THEN
          Wr.PutText(stderr, "ideal[" & Fmt.Pad(Fmt.Int(iIndex), 6) & "]\n");
          PZCandidate.Print(stderr, cand := iCand[iIndex])
        END
      END;

      (* Count matched found cands and print bogus ones: *)
      IF printBogus THEN Wr.PutText(stderr, "=== found NOT in ideal ===\n") END;
      fNOK := 0;
      FOR fIndex := 0 TO fN-1 DO
        IF fOK[fIndex] THEN
          INC(fNOK)
        ELSIF printBogus THEN
          Wr.PutText(stderr, "found [" & Fmt.Pad(Fmt.Int(fIndex), 6) & "]\n");
          PZCandidate.Print(stderr, cand := fCand[fIndex])
        END
      END;

      Wr.PutText(stderr, "\n");
      Wr.PutText(stderr, "ideal candidates =   " & Fmt.Int(iN) & "\n");
      Wr.PutText(stderr, "found candidates =   " & Fmt.Int(fN) & "\n");
      Wr.PutText(stderr, "ideal in found =     " & Fmt.Int(iNOK) & "\n");
      Wr.PutText(stderr, "ideal NOT in found = " & Fmt.Int(iN - iNOK) & "\n");
      Wr.PutText(stderr, "found in ideal =     " & Fmt.Int(fNOK) & "\n");
      Wr.PutText(stderr, "found NOT in ideal = " & Fmt.Int(fN - fNOK) & "\n");
      Wr.PutText(stderr,"\n");
      WITH
        sensitivity = FLOAT(iNOK,LONG)/FLOAT(iN,LONG),
        significance = FLOAT(iNOK,LONG)/FLOAT(fN,LONG),
        multiplicity = FLOAT(fNOK,LONG)/FLOAT(iNOK,LONG)
      DO
        Wr.PutText(stderr, "sensitivity =  " & FLR(sensitivity, 6,4) & "\n");
        Wr.PutText(stderr, "significance = " & FLR(significance, 6,4) & "\n");
        Wr.PutText(stderr, "multiplicity = " & FLR(multiplicity, 6,4) & "\n");
      END;
      Wr.PutText(stderr,"\n");

      (* Rearrange candidates to put the good/present ones in front: *)
      SortOK(iCand, iOK);
      SortOK(fCand, fOK);
    END;
  END ProcessCandidates;
  
PROCEDURE ProcessCandidatesOfCurvePair(
    iIni: NAT;                         (* Actual index of "iCand[0]" *)
    READONLY iCand: PZCandidate.List;  (* Ideal candidates of a curve pair. *)
    VAR (*IO*) iOK: BOOLS;             (* Identifies non-missing ideal candidates. *)
    fIni: NAT;                         (* Actual index of "fCand[0]" *)        
    READONLY fCand: PZCandidate.List;  (* Found candidates of a curve pair. *) 
    VAR (*IO*) fOK: BOOLS;             (* Identifies correct found candidates. *)
    minSteps: NAT;
    maxAdjust: NAT;
    printMatches: BOOL;
    VAR (*WORK*) ra, rb: REF PZMatch.T;
  ) =
  (*
    Processes a set of ideal and found candidates belonging to 
    the same curve pair. Sets "iOK[i]" when the ideal candidate
    "iCand[i]" is present in "fCand", and vice-versa.
  *)
  BEGIN
    FOR i := 0 TO LAST(iCand) DO
      WITH
        iC = iCand[i]
      DO
        FOR f := 0 TO LAST(fCand) DO
          WITH
            fC =  fCand[f],
            overlap = PZCandidate.Overlap(iC, fC, 
              minSteps := minSteps, maxAdjust := maxAdjust,
              ra := ra, rb := rb
            )
          DO
            iOK[i] := iOK[i] OR overlap;
            fOK[f] := fOK[f] OR overlap;
            IF overlap AND printMatches THEN
              Wr.PutText(stderr, "\n");
              PrintCandidate("ideal", i + iIni, iC);
              PrintCandidate("found", f + fIni, fC);
              Wr.PutText(stderr, "\n");
            END;
          END;
        END;
      END
    END
  END ProcessCandidatesOfCurvePair;
  
PROCEDURE PrintCandidate(tag: TEXT; i: NAT; READONLY c: PZCandidate.T) =
  BEGIN
    Wr.PutText(stderr, Fmt.Pad(tag,5) & "[" & Fmt.Pad(Fmt.Int(i), 6) & "]\n");
    PZCandidate.Print(stderr, cand := c);
  END PrintCandidate;

PROCEDURE WriteCandidateSubset(
    name: TEXT;                 (* File name prefix. *)
    VAR cand: PZCandidate.List; (* The candidate subset. *)
    desc: TEXT;                 (* Description of the subset. *)
    cmt: TEXT;                  (* Original dataset comments. *)
    lambda: LONG;               (* Filtering scale. *)
  ) =
  (*
    Writes the specified candidates to disk, as "fname.can",
    provided "fname" is not the empty string.
    NOTE: Will also reorder the candidates.  *)
  BEGIN
    IF NOT Text.Empty(name) THEN
      WITH
        wr = FileWr.Open(name & ".can"),
        cmt = cmt & "\n\nPZCompCands: " & desc
      DO
        PZCandidate.Sort(cand, PZCandidate.MismatchBetter);
        PZCandidate.Write(wr, cmt, cand, lambda)
      END
    END;
  END WriteCandidateSubset;

PROCEDURE FLR(x: LONG; w, p: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.LongReal(x, Fmt.Style.Fix, prec := p), w)
  END FLR;

PROCEDURE FI(x: NAT; w: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.Int(x), w)
  END FI;

<*UNUSED*>
PROCEDURE DebugIntPair(title: TEXT; r, s: INT) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, title);
    Wr.PutText(stderr, " ");
    Wr.PutText(stderr, Fmt.Int(r));
    Wr.PutText(stderr, " ");
    Wr.PutText(stderr, Fmt.Int(s));
    Wr.PutText(stderr, "\n");
  END DebugIntPair;

BEGIN
  Main()
END PZCompCands.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
