MODULE PZDraw EXPORTS Main;

(* Draws contours *)
(* Last edited on 1999-12-23 00:49:01 by hcgl *)

(*
  This program reads zero or more FloatChains from standard input,
  generates a Postscript (".ps" or ".eps") drawing of them. *)

IMPORT PSPlot, ParseParams, Process, FileRd, Wr, Rd, Lex, Fmt;
IMPORT FileFmt, NGet, FGet, OSError, Thread, Stdio, Text;
IMPORT PZLR3Chain, PZLRChain, PZGeo, PZPlot;
 
FROM PSPlot IMPORT Color, Invisible;
FROM Stdio IMPORT stderr, stdin;
FROM PZTypes IMPORT LONG, NAT, BOOL;
 
<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  Options = RECORD
      window: TEXT;        (* Plot window file name *)
      rect: TEXT;          (* Highlight rectangle file name *)
      plotStep: LONG;      (* Sampling step for plot *)
      output: TEXT;        (* Output file name, minus ".ps" *)
      epsFormat: BOOL;     (* TRUE to generate encapsulated Postscript *)
      ticks: BOOL;         (* TRUE to plot a tick mark every 0.05*NUMBER(chain) points. *)
      dots: BOOL;          (* TRUE to plot dots at positions given in "dotsFile". *)
      dotsFile : TEXT;     (* Name of dots file (a PZLR3Chain.T), incl. extension  *) 
      dotsPrec: NAT;       (* Formatting prec for 3rd column of "dotsFile"; 0=omit it. *) 
      grid: LONG;          (* Grid line spacing, or 0.0d0 if none. *)
      axes: BOOL;          (* TRUE to plot coordinate axes *)
      open: BOOL;          (* TRUE omits step from last sample to first sample. *)
      cubic: BOOL;         (* TRUE to use cubic interpolation for sampling *)
      
      (* Options that control the `filter scale' icon at corner: *)
      filterIconStyle: PZPlot.ScaleIconStyle;   (* Icon style. *)
      filterLambda: LONG;  (* Width (std. dev.) for gaussian graph, or 0 *)
      filterDotStep: LONG; (* Spacing of dots on Gaussian graph, or 0 *)
    END;
    
  Window = RECORD
      XMin, XMax, YMin, YMax : LONG;
    END;

PROCEDURE Main() =
  <* FATAL Rd.Failure *>
  VAR f:  PSPlot.File;
  BEGIN
    WITH
      o = GetOptions(),
      win = ReadWindow(FileRd.Open(o.window & ".wnd"))
    DO
      f :=  OpenPSFile(o.output, o.epsFormat, win);

      IF o.axes THEN
        f.setLineWidth(0.1);
        f.setLineSolid();
        f.setLineColor(Color{0.5, 0.5, 0.5});
        f.coordLine(PSPlot.Axis.X, 0.0d0);
        f.coordLine(PSPlot.Axis.Y, 0.0d0);
      END;

      IF o.grid > 0.0d0 THEN
        f.setLineWidth(0.1);
        f.setLineSolid();
        f.setLineColor(Color{0.5, 0.5, 0.5});
        FOR x := CEILING(win.XMin/o.grid) TO FLOOR(win.XMax/o.grid) DO
          f.coordLine(PSPlot.Axis.X, o.grid * FLOAT(x, LONG))
        END;
        FOR y := CEILING(win.YMin/o.grid) TO FLOOR(win.YMax/o.grid) DO
          f.coordLine(PSPlot.Axis.Y, o.grid * FLOAT(y, LONG))
        END;
      END;

      LOOP
        Lex.Skip(stdin);
        IF Rd.EOF(stdin) THEN EXIT END;
        WITH 
          a = PZLR3Chain.Read(stdin, headerOnly := FALSE),
          original = a.c^,
          sampled = SampleChain(a.c, o.plotStep, o.cubic)^
        DO
          IF NOT o.epsFormat THEN 
            NARROW(f, PSPlot.PSFile).caption(a.cmt)
          END;
          f.setLineSolid();
          f.setLineWidth(0.15);
          f.setLineColor(Color{0.0, 0.0, 0.0});
          f.setFillColor(Invisible);
          PZPlot.Curve(f, sampled, step := 1, closed := NOT o.open);
          IF o.ticks THEN
            PlotTickMarks(f, original)
          END
        END
      END;
      IF o.dots   THEN   
        WITH
          er = PZLR3Chain.Read(FileRd.Open(o.dotsFile), headerOnly := FALSE),
          e  = er.c^
        DO 
          PlotDotsMarks(f, e, o.dotsPrec);
        END;  
      END;

      IF NOT Text.Empty(o.rect) THEN
        WITH
          rec = ReadWindow(FileRd.Open(o.rect & ".wnd"))
        DO
          f.setLineSolid();
          f.setLineColor(Color{0.0, 0.0, 0.0});
          f.setFillColor(Invisible);
          f.setLineWidth(0.1);
          f.rectangle(rec.XMin, rec.XMax, rec.YMin, rec.YMax)
        END;
      END;

      IF o.filterLambda > 0.0d0 THEN
        PZPlot.ScaleIcon(f, 
          filterLambda := o.filterLambda, 
          filterDotStep := o.filterDotStep,
          style := o.filterIconStyle
        )
      END;
      ClosePSFile(f);
    END
  END Main;
  
PROCEDURE SampleChain(
    rp: REF PZLR3Chain.T; 
    step: LONG;
    cubic: BOOL;
  ): REF PZLR3Chain.T = 
  BEGIN
    IF step <= 0.0d0 THEN 
      RETURN rp
    ELSE
      WITH
        p = rp^,
        m = NUMBER(p),
        s = NEW(REF PZLRChain.T, m)^,
        length = PZLR3Chain.LinearLengths(p, s),
        n = CEILING(length/step),
        rq = NEW(REF PZLR3Chain.T, n), q = rq^
      DO
        IF cubic THEN
          WITH
            dp = NEW(REF PZLR3Chain.T, m)^,
            dq = NEW(REF PZLR3Chain.T, n)^
          DO
            PZLR3Chain.EstimateVelocities(s, length, p, dp, PZGeo.EstimateVelocityL);
            PZLR3Chain.HermiteUniformSample(s, length, p, dp, 0.0d0, q, dq)
          END
        ELSE
          PZLR3Chain.LinearUniformSample(s, length, p, 0.0d0, q)
        END;
        RETURN rq
      END
    END
  END SampleChain;
  
PROCEDURE PlotTickMarks(f: PSPlot.File; READONLY c: PZLR3Chain.T) =
  CONST NTicks = 20;
  BEGIN
    WITH 
      n = NUMBER(c)
    DO 
      f.setTextColor(Color{0.5, 0.5, 0.5});
      FOR k := 0 TO NTicks-1 DO 
        WITH 
          r = ROUND(FLOAT(k,REAL)/FLOAT(NTicks,REAL)*100.0),
          i = ROUND(FLOAT(k*n,REAL)/FLOAT(NTicks,REAL))
        DO
          f.label( 
            Fmt.Pad(Fmt.Int(r), 2, '0'),
            c[i][0],c[i][1]
          );
        END
      END;
      f.setLineColor(Color{0.0, 0.0, 0.0});
      FOR k := 0 TO NTicks-1 DO 
        WITH 
          i = ROUND(FLOAT(k*n,REAL)/FLOAT(NTicks,REAL))
        DO
          f.dot(c[i][0],c[i][1],2.0);
        END;
      END;
    END;  
  END PlotTickMarks;
  
PROCEDURE PlotDotsMarks(
    f: PSPlot.File; 
    READONLY e: PZLR3Chain.T; 
    prec : NAT 
  ) =
  BEGIN
    WITH 
      n = NUMBER(e)
    DO 
      f.setTextColor(Color{0.5, 0.5, 0.5});
      IF prec > 0 THEN 
        FOR i:=0 TO n-1 DO 
          f.label( 
              Fmt.Pad(Fmt.LongReal(e[i][2]), prec),
              e[i][0],e[i][1]
          );
        END;
      END;  
      f.setLineColor(Color{0.0, 0.0, 0.0});
      FOR i:=0 TO n-1 DO 
        f.dot(e[i][0],e[i][1],3.0);
      END;
    END;  
  END PlotDotsMarks;
  
PROCEDURE OpenPSFile(
    output: TEXT; 
    epsFormat: BOOL; 
    READONLY win: Window; 
  ): PSPlot.File =
  VAR f: PSPlot.File;
  CONST 
    PlotSize = 150.0d0;         (* Actual plot size in millimeters *)
  BEGIN
    WITH
      dX = win.XMax - win.XMin,
      dY = win.YMax - win.YMin,
      maxDim = MAX(dX, dY),
      plotXSize = dX/maxDim * PlotSize,
      plotYSize = dY/maxDim * PlotSize
    DO
      IF epsFormat THEN
        WITH
          name = output & ".eps",
          eps = NEW(PSPlot.EPSFile).open(name, xSize := plotXSize, ySize := plotYSize)
        DO
          f := eps
        END
      ELSE
        WITH
          name = output & ".ps",
          ps = NEW(PSPlot.PSFile).open(name)
        DO
          ps.beginPage();
          ps.beginDrawing(xSize := plotXSize, ySize := plotYSize);
          ps.setTextColor(Color{0.0, 0.0, 0.0});
          f := ps;
        END;
      END;
    END;
    f.setScale(PSPlot.Axis.X, win.XMin, win.XMax);
    f.setScale(PSPlot.Axis.Y, win.YMin, win.YMax);
    RETURN f
  END OpenPSFile;
  
PROCEDURE ClosePSFile(f: PSPlot.File) =
  BEGIN
    f.close()
  END ClosePSFile;
  
CONST 
  FileVersion = "96-04-11";

PROCEDURE ReadWindow(rd: Rd.T): Window =
  VAR w: Window;
  BEGIN
    FileFmt.ReadHeader(rd, "plot window", FileVersion);
    w.XMin := NGet.LongReal(rd, "XMin"); FGet.EOL(rd); 
    w.XMax := NGet.LongReal(rd, "XMax"); FGet.EOL(rd); 
    w.YMin := NGet.LongReal(rd, "YMin"); FGet.EOL(rd); 
    w.YMax := NGet.LongReal(rd, "YMax"); FGet.EOL(rd); 
    FileFmt.ReadFooter(rd, "plot window");
    RETURN w
  END ReadWindow;

<*UNUSED*>   
PROCEDURE ReadComments(rd: Rd.T): TEXT =
  <* FATAL Rd.Failure *>
  BEGIN
    RETURN Rd.GetText(rd, LAST(NAT))
  END ReadComments;
  
PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-window");
        o.window := pp.getNext();

        o.open := pp.keywordPresent("-open");
        
        IF pp.keywordPresent("-rect") THEN
          o.rect := pp.getNext()
        ELSE
          o.rect := ""
        END;

        IF pp.keywordPresent("-plotStep") THEN
          o.plotStep := pp.getNextLongReal(0.0d0);
          o.cubic := pp.keywordPresent("-cubic")
        ELSE
          o.plotStep := 0.0d0;
          o.cubic := FALSE;
        END;

        pp.getKeyword("-output");
        o.output := pp.getNext();

        o.ticks := pp.keywordPresent("-ticks");
        
        o.dots := pp.keywordPresent("-dots");
        IF o.dots  THEN
          o.dotsFile :=  pp.getNext(); 
          IF pp.keywordPresent("-prec") THEN
            o.dotsPrec := pp.getNextInt(0,16);
          ELSE 
            o.dotsPrec:= 0;
          END;
        END;
        
        o.axes := pp.keywordPresent("-axes");

        IF pp.keywordPresent("-grid") THEN
          o.grid := pp.getNextLongReal(0.0d0, LAST(LONG))
        ELSE
          o.grid := 0.0d0
        END;

        IF pp.keywordPresent("-filterLambda") THEN
          o.filterLambda := pp.getNextLongReal(0.0d0, LAST(LONG));
          IF pp.keywordPresent("-filterDotStep") THEN
            o.filterDotStep := pp.getNextLongReal(0.0d0, LAST(LONG));
          ELSE
            o.filterDotStep := 0.0d0
          END
        ELSE
          o.filterLambda := 0.0d0;
          o.filterDotStep := 0.0d0
        END;
        
        IF pp.keywordPresent("-filterIconStyle") THEN
          WITH s = pp.getNext() DO
            IF Text.Equal(s, "Circle") THEN
              o.filterIconStyle := PZPlot.ScaleIconStyle.Circle
            ELSIF Text.Equal(s, "Kernel") THEN
              o.filterIconStyle := PZPlot.ScaleIconStyle.Kernel
            ELSE
              pp.error("bad filter scale icon style")
            END
          END
        ELSE
          o.filterIconStyle := PZPlot.ScaleIconStyle.Circle
        END;

        o.epsFormat := pp.keywordPresent("-epsFormat"); 
               
        pp.finish();                                       
      EXCEPT                                                            
      | ParseParams.Error =>                                              
          Wr.PutText(stderr, "Usage: cat CURVEFILE.. \\\n");
          Wr.PutText(stderr, "  | PZDraw \\\n");
          Wr.PutText(stderr, "      -output PSNAME \\\n");
          Wr.PutText(stderr, "      -window WINDNAME \\\n");
          Wr.PutText(stderr, "      [ -open ] \\\n");
          Wr.PutText(stderr, "      [ -plotStep NUM [ -cubic ] ] \\\n");
          Wr.PutText(stderr, "      [ -axes ] [ -ticks | -dots <name> [ -prec NUM ] ] \\\n");
          Wr.PutText(stderr, "      [ -grid NUM ] \\\n");
          Wr.PutText(stderr, "      [ -filterLambda NUM [ -filterDotStep NUM ] ] \\\n");
          Wr.PutText(stderr, "      [ -filterIconStyle { Circle | Kernel } ] \\\n");
          Wr.PutText(stderr, "      [ -rect WINDNAME ] \\\n");
          Wr.PutText(stderr, "      [ -epsFormat ] \n");
        Process.Exit(1);
      END;
    END;
    RETURN o
    
  END GetOptions;
  

BEGIN
  Main()
END PZDraw.

(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
