MODULE PZDrawSegs EXPORTS Main;

(* Draws a list of chain segments   *)
(* Last edited on 1999-11-08 05:42:36 by hcgl *)

IMPORT PSPlot, LR3, LR4x4; 
IMPORT ParseParams, FileRd, Wr, Fmt;
IMPORT Process, OSError, Thread, Stdio;
IMPORT PZLR3Chain, PZSegment;
IMPORT PZPlot, PZGeo;  

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, NAT, BOOL;

<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  Options = RECORD
      input: TEXT;        (* Input segment file name *)
      chainDir: TEXT;     (* Directory where to find curve data *)
      chainPrefix: TEXT;  (* Invariant chain file name prefix *)
      band: NAT;          (* Nominal band width (lambda) for file names *)
      extension: TEXT;    (* Extension for geometric chains (".flc" usually) *)
      output: TEXT;       (* Output plot file (without ".ps"/".eps") *)
      maxSegs: NAT;       (* Num segments to plot *)
      
      (* Plotting options *)
      drawEvery: NAT;     (* Plot only one sample every this many. *) 
      epsFormat: BOOL;    (* TRUE to use Encapsulated Postscript. *)
      plotWidth: LONG;    (* Nominal segment plot width in mm. *)
      plotHeight: LONG;   (* Nominal segment plot width in mm. *)
      nSegsPerRow: NAT;   (* Number of segment plots per row (1 for ".eps"). *)
      nSegsPerCol: NAT;   (* Number of segment plots per column (1 for ".eps"). *)
    END;

PROCEDURE Main() =
  BEGIN
    WITH 
      o = GetOptions()
    DO
      IF o.maxSegs > 0 THEN
        WITH
          segData = PZSegment.Read(FileRd.Open(o.input & ".seg")),
          seg = segData.s^,
          curves = PZSegment.GetCurves(seg)^,
          chData = PZLR3Chain.ReadAll(
            o.chainPrefix, o.band, o.extension, 
            sel := curves, dir := o.chainDir, headerOnly := FALSE
          )
        DO
          DrawSegments(o.output, seg, chData.chain^, o);
        END;
      END;
    END;
 END Main;          

      
PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-input");
        o.input := pp.getNext();
        
        IF pp.keywordPresent("-chainDir") THEN
          o.chainDir := pp.getNext()
        ELSE
          o.chainDir := "."
        END;
        
        pp.getKeyword("-chainPrefix");
        o.chainPrefix := pp.getNext();
        
        pp.getKeyword("-band");
        o.band := pp.getNextInt();
          
        IF pp.keywordPresent("-extension") THEN 
          o.extension:= pp.getNext();
        ELSE 
          o.extension := ".flc";
        END;
        
        pp.getKeyword("-output");
        o.output := pp.getNext();
               
        IF pp.keywordPresent("-maxSegs") THEN
          o.maxSegs := pp.getNextInt(1)
        ELSE
          o.maxSegs := 160
        END;
        
        o.epsFormat := pp.keywordPresent("-epsFormat");
        
        IF o.epsFormat THEN
          o.nSegsPerRow := 1;
          o.nSegsPerCol := 1;
          IF pp.keywordPresent("-plotSize") THEN 
            o.plotWidth  := pp.getNextLongReal(10.0d0, 1000.0d0);
            o.plotHeight := pp.getNextLongReal(10.0d0, 1000.0d0);
          ELSE 
            o.plotWidth  := 40.0d0;
            o.plotHeight := 40.0d0;
          END;
        ELSE
          IF pp.keywordPresent("-nSegsPerPage") THEN 
            o.nSegsPerRow:= pp.getNextInt(1, 100);
            o.nSegsPerCol:= pp.getNextInt(1, 100);
          ELSE 
            o.nSegsPerRow := 4;
            o.nSegsPerCol := 5;
          END;
          o.plotWidth  := 160.0d0/FLOAT(o.nSegsPerRow, LONG);
          o.plotHeight := 200.0d0/FLOAT(o.nSegsPerCol, LONG);
        END;
        
        IF pp.keywordPresent("-drawEvery") THEN 
          o.drawEvery:= pp.getNextInt(1);
        ELSE 
          o.drawEvery := 1;
        END;
        
        pp.finish();   
        
      EXCEPT                                                            
      | ParseParams.Error =>                                              
          Wr.PutText(stderr, "Usage: PZDrawSegs \\\n");
          Wr.PutText(stderr, "  -input NAME \\\n");
          Wr.PutText(stderr, "  [ -chainDir DIR ] -chainPrefix NAME \\\n");
          Wr.PutText(stderr, "  -band NUMBER \\\n");
          Wr.PutText(stderr, "  [ -extension EXT ] \\\n");
          Wr.PutText(stderr, "  -output NAME \\\n");
          Wr.PutText(stderr, "  [ -maxSegs NUMBER ] \\\n");
          Wr.PutText(stderr, "  [ -epsFormat [ -plotSize WIDTH HEIGHT ] | \\\n");
          Wr.PutText(stderr, "    -nSegsPerPage NUMBER NUMBER \\\n");
          Wr.PutText(stderr, "  ] \\\n");
          Wr.PutText(stderr, "  [ -drawEvery NUMBER ] \n");
        Process.Exit(1);
      END;
    END;
    RETURN o
  END GetOptions;

PROCEDURE DrawSegments(
    output: TEXT;
    READONLY seg: PZSegment.List;
    READONLY flc: ARRAY OF PZLR3Chain.ReadData;
    READONLY o: Options;
  )=
  VAR ps: PSPlot.File;
  BEGIN 
    Wr.PutText(stderr,"DrawSegments...\n");
    WITH 
      nSegs = MIN(o.maxSegs, NUMBER(seg)),
      nSegsPerPage = o.nSegsPerRow * o.nSegsPerCol,
      nPages = (nSegs + nSegsPerPage - 1) DIV nSegsPerPage
    DO
      IF NOT o.epsFormat THEN
        WITH name = output & ".ps" DO
          ps := NEW(PSPlot.PSFile).open(name);
          Wr.PutText(stderr, "writing " & name & "\n");  
        END
      END;
    
      FOR page := 0 TO nPages-1  DO 
        WITH
          firstSegInPage = page*nSegsPerPage,
          lastSegInPage = MIN(nSegs, (page+1)*nSegsPerPage) - 1,
          nSegsInPage = lastSegInPage - firstSegInPage + 1,
          caption = "Segments [" & 
            Fmt.Int(firstSegInPage) & ".." & Fmt.Int(lastSegInPage) &
          "]"
        DO
          IF o.epsFormat THEN
            <* ASSERT firstSegInPage = lastSegInPage *>
            WITH 
              pg = Fmt.Pad(Fmt.Int(page), 5, '0'),
              name = output & "-pg" & pg & ".eps",
              xSize = o.plotWidth,
              ySize = o.plotHeight
            DO
              ps := NEW(PSPlot.EPSFile).open(name, xSize := xSize, ySize := ySize);
              Wr.PutText(stderr, "writing " & name & "\n");  
            END
          ELSE
            WITH ps = NARROW(ps, PSPlot.PSFile) DO
              ps.beginPage();
              (* ps.caption(output); *)
              (* ps.caption(caption); *) 
            END;
          END;
          Wr.PutText(stderr, caption & "\n");
          DrawPage(ps, SUBARRAY(seg, firstSegInPage, nSegsInPage), flc, o);
          IF o.epsFormat THEN 
            ps.close()
          ELSE
            WITH ps = NARROW(ps, PSPlot.PSFile) DO ps.endPage() END;
          END;
        END;
      END;
      IF NOT o.epsFormat THEN ps.close() END;
    END   
  END DrawSegments;
   
PROCEDURE DrawPage(
    ps: PSPlot.File;
    READONLY seg: PZSegment.List;
    READONLY flc: ARRAY OF PZLR3Chain.ReadData;
    READONLY o: Options;
  )=
  CONST NominalSegmentWidth = 750.0d0;
  BEGIN 
    FOR i := 0 TO LAST(seg)  DO 
      IF NOT o.epsFormat THEN
        BeginDrawing(ps, o.nSegsPerRow, o.nSegsPerCol, o.plotWidth, o.plotHeight, i)
      END; 
      WITH 
        aspect = o.plotHeight/o.plotWidth,
        xMax = NominalSegmentWidth/2.0d0,
        yMax = NominalSegmentWidth/2.0d0*aspect
      DO
        ps.setScale(PSPlot.Axis.X, -xMax, +xMax);
        ps.setScale(PSPlot.Axis.Y, -yMax, +yMax)
      END;
      ps.setLineWidth(0.2);
      ps.setLineColor(PSPlot.Black);
      ps.frame();
      DrawSegment(ps, seg[i], flc, o);
      IF NOT o.epsFormat THEN EndDrawing(ps) END;
    END   
  END DrawPage;

  
PROCEDURE BeginDrawing(
    ps: PSPlot.PSFile; 
    nSegsPerRow: NAT; 
    nSegsPerCol: NAT; 
    plotWidth: LONG;
    plotHeight: LONG;
    i: NAT;
  ) =
  BEGIN
    WITH 
      totWidth = FLOAT(nSegsPerRow, LONG) * plotWidth,
      plotXMargin = 0.5d0 * (PSPlot.LetterXSize - totWidth),
      plotYMargin = plotXMargin,
      plotX = i MOD nSegsPerRow,
      plotY = nSegsPerCol - 1 - (i DIV nSegsPerRow),
      xCenter = plotWidth  * (FLOAT(plotX, LONG) + 0.5d0) + plotXMargin,
      yCenter = plotHeight * (FLOAT(plotY, LONG) + 0.5d0) + plotYMargin
    DO
      ps.beginDrawing(plotWidth, plotHeight, xCenter, yCenter)
    END;
  END BeginDrawing;
 
   
PROCEDURE EndDrawing(ps: PSPlot.PSFile) =
  BEGIN
    ps.endDrawing()
  END EndDrawing;
 
   
PROCEDURE DrawSegment(
    READONLY ps: PSPlot.File;
    READONLY s: PZSegment.T;
    READONLY flc: ARRAY OF PZLR3Chain.ReadData;
    READONLY o: Options;
  )=
  VAR 
    w: LR3.T := LR3.T{-1.0d0, 0.0d0, 0.0d0}; 
    ini, fin: NAT;
  BEGIN
    WITH 
      chainNum = s.cvx,
      chain = flc[chainNum].c^
    DO
      <* ASSERT NUMBER(chain) = s.tot *>
      ini := s.ini MOD s.tot;
      fin := (ini + s.ns - 1) MOD s.tot;
      IF s.ns = 1 THEN 
        <* ASSERT s.tot >= 4 *>
        ini := (ini-1) MOD s.tot;
        fin := (fin+1) MOD s.tot
      END;  
      WITH 
        p = chain[ini],
        q = chain[fin],
        ctr = PZGeo.SegMid(p, q), 
        u = PZGeo.SegDir(p, ctr),
        disp = (LR3.T){-ctr[0],-ctr[1],-ctr[2]},
        t = PZGeo.Translation(disp), (* -ctr *)
        r = PZGeo.Rotation(u, w),
        m = LR4x4.Mul(t,r),
        mappedChain = PZLR3Chain.Map(chain, m)^
      DO 
        PZPlot.HalfCandidate(
          ps, s, mappedChain, color := PSPlot.Red,
          whole := TRUE, closed := TRUE, 
          dots := FALSE, pointers := TRUE, label := TRUE,
          step := o.drawEvery
        );
      END;
    END
  END DrawSegment;

 
BEGIN
  Main()
END PZDrawSegs.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
