MODULE PZMapSegs EXPORTS Main;
(* Last edited on 1999-11-08 05:42:01 by hcgl *)

(* Maps a list of curve segments to another scale *)

IMPORT ParseParams, Process, FileRd, FileWr;
IMPORT Wr, OSError, Thread, Stdio, Fmt, TextWr;
IMPORT PZLRChain, PZSymbolChain;
IMPORT PZSegment;

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, NAT;

<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  Options = RECORD
      input: TEXT;        (* Input segment file name. *)
      inBand: NAT;        (* Input nominal filtering bandwidth. *)
      chainDir: TEXT;     (* Directory where to find curve data *)
      chainPrefix: TEXT;  (* Curve file name prefix. *)
      output: TEXT;       (* Output segment file name. *)
      outBand: NAT;       (* Output nominal filtering bandwidth. *)
      outStep: LONG;      (* Sampling step of output curves. *)
      blurFactor: LONG;   (* Corner broadening factor (may be negative). *)
      extraLength: LONG;  (* Extra length to add at each segment end. *)
    END;

PROCEDURE Main() =
  BEGIN
    WITH
      o = GetOptions(),
      sData =   PZSegment.Read(FileRd.Open(o.input & ".seg")),
      oldSeg =  sData.s^,
      curves =  PZSegment.GetCurves(oldSeg)^,
      oldLabel = PZLRChain.ReadAll(
        o.chainPrefix, o.inBand, ".lbl", sel := curves, dir := o.chainDir, 
        headerOnly := FALSE
      ),
      newLabel = PZLRChain.ReadAll(
        o.chainPrefix, o.outBand, ".lbl", sel := curves, dir := o.chainDir,
        headerOnly := FALSE
      ),
      newCvc =  PZSymbolChain.ReadAll(
        o.chainPrefix, o.outBand, ".cvc", sel := curves, dir := o.chainDir,
        headerOnly := TRUE
      ),
      newSeg = MapAllSegments(
        oldSeg,
        oldLabel.chain^,
        newLabel.chain^, newCvc.chain^,
        o.inBand, o.outBand,
        newStep := o.outStep,
        blurFactor := o.blurFactor,
        extraLength := o.extraLength
      )^,
      wr = FileWr.Open(o.output & ".seg")
    DO
      PZSegment.Write(wr, sData.cmt & MapComments(o), newSeg);
    END;
  END Main;

PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        IF pp.keywordPresent("-chainDir") THEN
          o.chainDir := pp.getNext()
        ELSE
          o.chainDir := "."
        END;

        pp.getKeyword("-chainPrefix");
        o.chainPrefix := pp.getNext();

        pp.getKeyword("-input");
        o.input := pp.getNext();

        pp.getKeyword("-inBand");
        o.inBand := pp.getNextInt();

        pp.getKeyword("-output");
        o.output := pp.getNext();

        pp.getKeyword("-outBand");
        o.outBand := pp.getNextInt();

        pp.getKeyword("-outStep");
        o.outStep := pp.getNextLongReal(0.0d0, 1024.0d0);

        pp.getKeyword("-blurFactor");
        o.blurFactor := pp.getNextLongReal(-5.0d0, 5.0d0);

        pp.getKeyword("-extraLength");
        o.extraLength := pp.getNextLongReal(0.0d0);

        pp.finish();

      EXCEPT
      | ParseParams.Error =>
          Wr.PutText(stderr, "Usage: PZMapSegs \\\n");
          Wr.PutText(stderr, "  -input NAME \\\n");
          Wr.PutText(stderr, "  -inBand NUMBER \\\n");
          Wr.PutText(stderr, "  [ -chainDir DIR ] -chainPrefix   NAME \\\n");
          Wr.PutText(stderr, "  -output  NAME \\\n");
          Wr.PutText(stderr, "  -outBand  NUMBER \\\n");
          Wr.PutText(stderr, "  -outStep NUMBER \\\n");
          Wr.PutText(stderr, "  -blurFactor NUMBER \\\n");
          Wr.PutText(stderr, "  -extraLength NUMBER \n");
        Process.Exit(1);
      END;
    END;
    RETURN o
  END GetOptions;

PROCEDURE MapAllSegments(
    READONLY seg: PZSegment.List;
    READONLY oldLabel: ARRAY OF PZLRChain.ReadData;
    READONLY newLabel: ARRAY OF PZLRChain.ReadData;
    READONLY newCvc: ARRAY  OF PZSymbolChain.ReadData;
    oldBand:  NAT;
    newBand: NAT;
    newStep: LONG;
    blurFactor: LONG;
    extraLength: LONG;
  ): REF PZSegment.List =
  VAR
    nSegNew: NAT := 0;
  BEGIN
    WITH
      nSegOld = NUMBER(seg),
      segNew = NEW(REF PZSegment.List, nSegOld)^,
      halfShrink = blurFactor * FLOAT(newBand - oldBand, LONG) / newStep,
      halfStretch = extraLength / newStep,
      extra = ROUND(halfStretch - halfShrink)
    DO
      Wr.PutText(stderr, "input segments = " & Fmt.Int(nSegOld) & "\n");
      FOR i := 0 TO nSegOld-1  DO
        WITH
          sOld = seg[i],
          mappable = newLabel[sOld.cvx].c # NIL AND newCvc[sOld.cvx].samples # 0
        DO
          <* ASSERT oldLabel[sOld.cvx].stride = newLabel[sOld.cvx].stride *>
          IF mappable THEN
            WITH
              stride = oldLabel[sOld.cvx].stride,
              tOld = oldLabel[sOld.cvx].c^,
              tNew = newLabel[sOld.cvx].c^,
              sRaw = PZLRChain.MapSegment(sOld, tOld, tNew, stride),
              sNew = PZSegment.Expand(sRaw, extra, extra)
            DO
              IF sNew.ns > 1 THEN
                segNew[nSegNew] := sNew; INC(nSegNew)
              END
            END
          END;
        END
      END;
      Wr.PutText(stderr, "output segments = " & Fmt.Int(nSegNew) & "\n");
      WITH r = NEW(REF PZSegment.List, nSegNew) DO
        r^ := SUBARRAY(segNew, 0, nSegNew);
        RETURN r;
      END;
    END;
  END MapAllSegments;

PROCEDURE MapComments(READONLY o: Options):TEXT=
  BEGIN
    WITH
      wr = NEW(TextWr.T).init()
    DO
      Wr.PutText( wr, "PZMapSegs\n");
      Wr.PutText( wr, "  input =       \"" & o.input & "\"\n");
      Wr.PutText( wr, "  inBand =      " & Fmt.Int(o.inBand) & "\n");
      Wr.PutText( wr, "  chainDir =    \"" & o.chainDir     & "\"\n");
      Wr.PutText( wr, "  chainPrefix = \"" & o.chainPrefix     & "\"\n");
      Wr.PutText( wr, "  output =      \"" & o.output & "\n");
      Wr.PutText( wr, "  outBand =     " & Fmt.Int(o.outBand) & "\n");
      Wr.PutText( wr, "  blurFactor =  " & FLR(o.blurFactor,8,6) & "\n");
      Wr.PutText( wr, "  extraLength = " & FLR(o.extraLength,8,4) & "\n");
      RETURN(TextWr.ToText(wr))
    END (* DO *);
  END MapComments;

PROCEDURE FLR(x: LONG; w, p: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.LongReal(x, Fmt.Style.Fix, prec := p), w)
  END FLR;

BEGIN
  Main()
END PZMapSegs.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
