MODULE PZPlotMatchCost EXPORTS Main;

(* Plots graphs of symbol chain mismatch as a function of chain length *)
(* Last edited on 1999-11-08 05:36:15 by hcgl *)

IMPORT Wr, Fmt, Thread, OSError, Stdio, ParseParams;
IMPORT Process, Random, FileRd, FileWr;
IMPORT PZSymbolChain, PZSymbol, PZMatch, PZCandidate, PZMismatch;

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, LONGS, INT, NAT, BOOL;

<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  MismatchOptions = PZMismatch.Options;

  Options = RECORD
      input: TEXT;         (* Input candidate file (without extensions). *)
      maxCands: NAT;       (* Max number of candidates to consider. *)
      chainDir: TEXT;      (* Directory where chains reside. *)
      chainPrefix : TEXT;  (* Input chain file name prefix. *)
      band: NAT;           (* Nominal band width (used for chain file names). *)
      step: LONG;          (* Nominal sampling step. *)
      output: TEXT;        (* Output plot file name (without extensions). *)
      perturb: LONG;       (* Amount of perturbation to second chain. *)
      shift: NAT;          (* Amount of shifting to apply to second chain. *)
      maxShift: LONG;      (* Max shift to try in alignment. *)
      mp: MismatchOptions; (* Mismatch formula parameters. *)
      dontReverse: BOOL;   (* TRUE will not reverse the sense of chain 2. *)
      printMatches: BOOL;  (* TRUE to print matches. *)
    END;

TYPE
  Chains = ARRAY OF PZSymbolChain.ReadData;

  VTable = PZSymbol.Table;

PROCEDURE Main() =
  BEGIN
    WITH
      o = GetOptions()
    DO
      WITH
        candData =  PZCandidate.Read(FileRd.Open(o.input & ".can")),
        cand = candData.c^,
        nCands = NUMBER(cand),
        sel = PZCandidate.GetCurves(cand)^,
        chData = PZSymbolChain.ReadAll(
          o.chainPrefix, o.band, sel := sel, extension := ".cvc", 
          dir := o.chainDir, headerOnly := FALSE
        ),
        plotWr = FileWr.Open(o.output & ".plt"),
        dch = PZSymbol.MakeDecodeTable(chData.sigmaMin)^,
        ech = PZSymbol.MakeErrorVarTable(chData.sigmaMin)^
      DO
        <* ASSERT chData.sigmaMin = chData.sigmaMax *>
        Wr.PutText(stderr, Fmt.Int(nCands) & " candidates\n");
        FOR i := 0 TO MIN(o.maxCands, nCands) - 1 DO
          Wr.PutText(stderr, "candidate " & Fmt.Int(i) & "\n");
          ProcessCandidate(plotWr, cand[i], chData.chain^, dch, ech, o)
        END
      END
    END
  END Main;

PROCEDURE ProcessCandidate(
    plotWr: Wr.T;
    READONLY c: PZCandidate.T;
    READONLY chain: Chains;
    READONLY dch: VTable;
    READONLY ech: VTable;
    READONLY o: Options;
  ) =
  VAR minAvgDist: REF LONGS;
  BEGIN
    WITH
      ka  = c.seg[0].cvx,
      ca = chain[ka].c^,
      ia =  c.seg[0].ini,
      na =  c.seg[0].ns,
      fa = ia + na - 1,

      kb  = c.seg[1].cvx,
      cb = chain[kb].c^,
      ib =  c.seg[1].ini,
      nb = c.seg[1].ns,
      fb = ib + nb - 1,
      maxSteps = (na-1) + (nb-1)
    DO
      PZCandidate.Print(stderr, c);
      IF na > 3 AND nb > 3 THEN
        <* ASSERT ia < fa *>
        <* ASSERT ib < fb *>
        <* ASSERT o.mp.maxDistSqr < LAST(LONG) *>
        
        (* Compute best matches per length for all shifts: *)
        MatchSymbolicChains(o,
          ca, ia, fa, cb, ib, fb,
          dch := dch, ech := ech,
          (*OUT*)
          minAvgDist := minAvgDist
        );
        FOR nSteps := 2 TO maxSteps BY 2 DO
          WritePlotEntry(plotWr,
            ka, ia, fa, kb, ib, fb, 
            o.step, nSteps, minAvgDist[nSteps]
          )
        END;
        WritePlotBreak(plotWr)
      ELSE
        Wr.PutText(stderr, "too small, skipped\n");
      END
    END
  END ProcessCandidate;

PROCEDURE WritePlotEntry(
    plotWr: Wr.T;
    ka, ia, fa: NAT;
    kb, ib, fb: NAT;
    step: LONG;
    nSteps: NAT;
    minAvgDist: LONG;
  ) =
  BEGIN
    WITH
      length = step * FLOAT(nSteps, LONG) / 2.0d0
    DO
      Wr.PutText(plotWr, Fmt.Pad(Fmt.Int(ka), 4, '0'));
      Wr.PutText(plotWr, " ");
      Wr.PutText(plotWr, Fmt.Pad(Fmt.Int(ia), 4));
      Wr.PutText(plotWr, " ");
      Wr.PutText(plotWr, Fmt.Pad(Fmt.Int(fa), 4));
      Wr.PutText(plotWr, " ");
      Wr.PutText(plotWr, Fmt.Pad(Fmt.Int(kb), 4, '0'));
      Wr.PutText(plotWr, " ");
      Wr.PutText(plotWr, Fmt.Pad(Fmt.Int(ib), 4));
      Wr.PutText(plotWr, " ");
      Wr.PutText(plotWr, Fmt.Pad(Fmt.Int(fb), 4));
      Wr.PutText(plotWr, " ");
      Wr.PutText(plotWr, FLR(length, 8,2));
      Wr.PutText(plotWr, "  ");
      Wr.PutText(plotWr, Fmt.Pad(Fmt.Int(nSteps), 8));
      Wr.PutText(plotWr, "  ");
      Wr.PutText(plotWr, FLR(minAvgDist, 8,6));
      Wr.PutText(plotWr, "\n");
    END;
    Wr.Flush(plotWr);
  END WritePlotEntry;

PROCEDURE WritePlotBreak(plotWr: Wr.T) =
  BEGIN
    Wr.PutText(plotWr, "\n");
  END WritePlotBreak;

PROCEDURE MatchSymbolicChains(
    READONLY o: Options;
    READONLY ca: PZSymbolChain.T;
    ia, fa: NAT;
    READONLY cb: PZSymbolChain.T;
    ib, fb: NAT;
    READONLY dch: VTable;
    READONLY ech: VTable;
    VAR (*OUT*) minAvgDist: REF LONGS;
  )=
  VAR
    mismatch: LONG;
    matchedLength: LONG;
    length: LONG;
    mt: REF PZMatch.T;
    rCost: REF PZMatch.CostMatrix := NIL;
  BEGIN
    WITH
      na = fa - ia + 1,
      a = PZSymbolChain.Trim(ca, ia, na)^,
      ma = NUMBER(a),
      nb = fb - ib + 1,
      b = PZSymbolChain.Trim(cb, ib, nb)^,
      mb = NUMBER(b),
      ac = (na-1) DIV 2,
      bc = (nb-1) DIV 2,
      rnd = NEW(Random.Default).init(TRUE),
      mtOld = PZMatch.T{PZMatch.Pair{ac,bc}}
    DO
      <* ASSERT na >= 2 *>
      <* ASSERT nb >= 2 *>
      IF NOT o.dontReverse THEN PZSymbolChain.ReverseAndComplement(b) END;
      IF o.perturb # 0.0d0 THEN
        FOR i := 0 TO nb-1 DO
          WITH perturb = ROUND(o.perturb * rnd.longreal(-1.0d0, +1.0d0)) DO
            b[i] := PZSymbol.FromInt(PZSymbol.ToInt(b[i]) + perturb)
          END
        END;
      END;
      IF o.shift # 0 THEN
        FOR i := 0 TO nb-1-o.shift DO b[i] := b[i+o.shift] END;
        FOR i := nb-o.shift TO nb-1 DO b[i] := b[i-1] END;
      END;
      mt := NIL;
      IF minAvgDist = NIL OR NUMBER(minAvgDist^) < ma+mb-1 THEN
        minAvgDist := NEW(REF LONGS, ma+mb-1)
      END;
      PZSymbolChain.RefineMatch(
        a, b,
        decode := dch, errorVar := ech,
        step := o.step,
        mtOld := mtOld,
        maxAdjust := CEILING(o.maxShift / o.step),
        maxDistSqr := o.mp.maxDistSqr,
        critDistSqr := o.mp.critDistSqr,
        skipDistSqr := o.mp.skipDistSqr,
        (*OUT*)
        mt := mt,
        mismatch := mismatch,
        length := length,
        matchedLength := matchedLength,
        minAvgDisc := SUBARRAY(minAvgDist^, 0, ma+mb-1),
        (*WORK*)
        rCost := rCost
      );
      IF o.printMatches THEN
        PrintSymbolicMatch(a, ia, fa, b, ib, fb, mt^, mismatch, length, matchedLength)
      END;
    END;
  END MatchSymbolicChains;

PROCEDURE PrintSymbolicMatch(
    READONLY a: PZSymbolChain.T;
    ia, fa: NAT;
    READONLY b: PZSymbolChain.T;
    ib, fb: NAT;
    READONLY mt: PZMatch.T;
    mismatch: LONG;
    length: LONG;
    matchedLength: LONG;
  ) =

  PROCEDURE Chain(j, i: NAT): CHAR =
    BEGIN
      IF j = 0 THEN RETURN a[i] ELSE RETURN b[i] END
    END Chain;

  BEGIN
    WITH
      nm = NUMBER(mt)
    DO
      Wr.PutText(stderr, "\n");
      (* Write matching portions of given chains, unaligned: *)
      FOR j := 0 TO 1 DO
        WITH
          ig = ARRAY [0..1] OF NAT{ia, ib}[j],
          fg = ARRAY [0..1] OF NAT{fa, fb}[j],
          im = mt[0][j], fm = mt[nm-1][j],
          nLetters = fm - im + 1
        DO
          Wr.PutText(stderr, "chain " & Fmt.Int(j) & ":");
          Wr.PutText(stderr, " given [" & FI(ig,4) & ".." & FI(fg,4) & "]");
          Wr.PutText(stderr, " (" & FI(fg - ig, 4) & ")");
          Wr.PutText(stderr, " matched [" & FI(im, 4) & ".." & FI(fm, 4) & "]");
          Wr.PutText(stderr, " (" & FI(fm - im, 4) & ")");
          Wr.PutText(stderr, " ");
          FOR r := 0 TO nLetters-1  DO
            WITH ch = Chain(j, im + r) DO
              Wr.PutText(stderr, Fmt.Char(ch))
            END;
          END;
          Wr.PutText(stderr, "\n");
        END;
      END;

      Wr.PutText(stderr, "\n");

      (* Write matched matching portions of given chains, aligned: *)
      FOR j := 0 TO 1 DO
        FOR r := 0 TO nm-1  DO
          WITH ch = Chain(j, mt[r][j]) DO
            Wr.PutText(stderr, Fmt.Char(ch))
          END;
        END;
        Wr.PutText(stderr, "\n");
      END;
      (* Write letter distances for each matched pair: *)
      FOR r := 0 TO nm-1  DO
        WITH
          cha = Chain(0, mt[r][0]),
          chb = Chain(1, mt[r][1]),
          d = ABS(PZSymbol.ToInt(cha) - PZSymbol.ToInt(chb))
        DO
          IF d > 9 THEN
            Wr.PutChar(stderr, '*')
          ELSE
            Wr.PutChar(stderr, VAL(ORD('0') + d, CHAR))
          END
        END;
      END;
      Wr.PutText(stderr, "\n");
      (* Print statistics: *)
      Wr.PutText(stderr, "mismatch = " & FLR(mismatch, 8,4) & "\n");
      Wr.PutText(stderr, "length   = " & FLR(length, 8,4) & "\n");
      Wr.PutText(stderr, "matched  = " & FLR(matchedLength, 8,4) & "\n");
      Wr.PutText(stderr, "\n");
    END
  END PrintSymbolicMatch;

PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-input");
        o.input := pp.getNext();

        IF pp.keywordPresent("-maxCands") THEN
          o.maxCands := pp.getNextInt(1, LAST(NAT))
        ELSE
          o.maxCands := LAST(NAT)
        END;

        IF pp.keywordPresent("-chainDir") THEN
          o.chainDir := pp.getNext()
        ELSE
          o.chainDir := "."
        END;
        
        pp.getKeyword("-chainPrefix");
        o.chainPrefix := pp.getNext();

        IF pp.keywordPresent("-perturb") THEN
          o.perturb := pp.getNextLongReal(0.0d0);
        ELSE
          o.perturb := 0.0d0;
        END;

        IF pp.keywordPresent("-shift") THEN
          o.shift := pp.getNextInt(0, 999);
        ELSE
          o.shift := 0
        END;

        pp.getKeyword("-output");
        o.output := pp.getNext();

        pp.getKeyword("-band");
        o.band := pp.getNextInt();

        pp.getKeyword("-step");
        o.step := pp.getNextLongReal(0.0d0);

        o.dontReverse := pp.keywordPresent("-dontReverse");

        o.mp := PZMismatch.ParseOptions(pp);

        pp.getKeyword("-maxShift");
        o.maxShift  := pp.getNextLongReal(0.0d0);

        o.printMatches := pp.keywordPresent("-printMatches");

        pp.finish();
      EXCEPT
      | ParseParams.Error =>
          Wr.PutText(stderr,  "Usage: PZPlotMatchCost \\\n");
          Wr.PutText(stderr,  "   -input FNAME \\\n");
          Wr.PutText(stderr,  "   -output FNAME \\\n");
          Wr.PutText(stderr,  "   [ -chainDir DIR ] -chainPrefix  NAME \\\n");
          Wr.PutText(stderr,  "   -band  NAT \\\n");
          Wr.PutText(stderr,  "   -maxShift NSTEPS \\\n");
          Wr.PutText(stderr,  "   " & PZMismatch.OptionsHelp & " \\\n");
          Wr.PutText(stderr,  "   [ -perturb DIST ] [ -shift NUM ]  \\\n");
          Wr.PutText(stderr,  "   [ -printMatches ] \n");
          Process.Exit(1);
      END;
    END;
    RETURN o
  END  GetOptions;

<*UNUSED*>
PROCEDURE DebugInt(text : TEXT; x, y, z: INT )=
  BEGIN
    Wr.PutText(stderr, text);
    Wr.PutText(stderr, "   " & Fmt.Int(x) & "   " & Fmt.Int(y)& "   " & Fmt.Int(z));
    Wr.PutText(stderr, "\n");
  END DebugInt;

PROCEDURE FLR(x: LONG; w, p: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.LongReal(x, Fmt.Style.Fix, prec := p), w)
  END FLR;

PROCEDURE FI(x: INT; w: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.Int(x), w)
  END FI;

BEGIN
  Main()
END PZPlotMatchCost.

(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
