MODULE PZRefineCandsGeometry EXPORTS Main;

(* Refines a list of candidate matchings, using DP on geometric chains.  *)
(* Last edited on 1999-12-22 16:35:01 by hcgl *)

IMPORT Process, OSError, Thread, ParseParams;
IMPORT FileRd, FileWr, TextWr, Wr, Fmt, Math;
IMPORT LR3, LR4x4;
IMPORT PZGeo, PZOptMatrix, PZFullPlot;
IMPORT PZLR3Chain, PZCandidate, PZMismatch, PZMatch, PZSegment;

FROM Math IMPORT sqrt;
FROM LR3 IMPORT Dist;
FROM LR4x4 IMPORT Mul, Identity;
FROM Minimizer IMPORT Value;

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, LONGS, INT, NAT, BOOL;

<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  MismatchOptions = PZMismatch.Options;
  Options = RECORD
      input: TEXT;         (* Input candidate file name. *)
      maxCands: NAT;       (* Max num of candidates to consider. *)
      chainDir: TEXT;      (* Directory where to find chain files. *)
      chainPrefix: TEXT;   (* Invariant chain file name prefix. *)
      band: NAT;           (* Nominal band width (lambda) for file names. *)
      step: LONG;          (* Sampling step. *)
      extension: TEXT;     (* Extension of chain files, usually ".flc" *)
      output: TEXT;        (* Output file name (without extensions) *)

      (* Matching parameters: *)
      simpleMatch: BOOL;   (* TRUE uses "FindBest" match instead of "Refine". *)
      maxShift: LONG;      (* Max shift to try in alignment. *)
      mp: MismatchOptions; (* Mismatch formula parameters. *)
      minLength: LONG;     (* Minimum length of matching segments. *)
      blurFactor: LONG;    (* Corner broadening factor. *)
      extraLength: LONG;   (* Extra length to add at each candidate end. *)

      (* Sort/prune parameters: *)
      maxCurveCands: NAT;  (* Max candidates per curve. *)
      maxPairCands: NAT;   (* Max candidates per curve pair. *)

      (* Placement/optimization parameters: *)
      dontGuess : BOOL;    (* TRUE to use identity as the initial matrix  *)
      perturbMatrix: BOOL; (* TRUE to perturb placement matrix *)
      maxEvals: NAT;       (* Max trials in evaluation (0 = dont' optimize). *)
      plotFunction: BOOL;  (* TRUE to plot the goal function. *)

      (* Plot parameters: *)
      draw: BOOL;          (* TRUE to draw the the initial and final placements *)
      epsFormat: BOOL;     (* TRUE: eps format, FALSE: plain ps format *)
      drawEvery: NAT;      (* Plot one every this many samples. *)
      displace: LONG;      (* Separate chains by this amount when plotting *)
    END;

TYPE
  ChainData = PZLR3Chain.ReadData;

PROCEDURE Main() =
  BEGIN
    WITH
      o = GetOptions(),
      candData = PZCandidate.Read(FileRd.Open(o.input & ".can")),
      oldCand = SUBARRAY(candData.c^, 0, MIN(o.maxCands, NUMBER(candData.c^))),
      lambda = candData.lambda,
      curves = PZCandidate.GetCurves(oldCand)^,
      chData = PZLR3Chain.ReadAll(
        o.chainPrefix, o.band, o.extension, sel := curves,
        dir := o.chainDir, headerOnly := FALSE
      ),
      newCand = RefineCandidates(oldCand, chData.chain^, o, lambda)^,
      runCmt = RunComments(o)
    DO
      WITH
        wr = FileWr.Open(o.output & ".can")
      DO
        PZCandidate.Write(wr, candData.cmt & "\n" & runCmt, newCand, lambda);
      END;
    END;
  END Main;

PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-input");
        o.input := pp.getNext();

        IF pp.keywordPresent("-maxCands") THEN
          o.maxCands := pp.getNextInt(1)
        ELSE
          o.maxCands := LAST(NAT)
        END;

        IF pp.keywordPresent("-chainDir") THEN
          o.chainDir := pp.getNext()
        ELSE
          o.chainDir := "."
        END;

        pp.getKeyword("-chainPrefix");
        o.chainPrefix := pp.getNext();

        pp.getKeyword("-band");
        o.band := pp.getNextInt();

        pp.getKeyword("-step");
        o.step := pp.getNextLongReal(0.0d0, 1024.0d0);

        IF pp.keywordPresent("-extension") THEN
          o.extension := pp.getNext()
        ELSE
          o.extension := ".flc"
        END;

        pp.getKeyword("-output");
        o.output := pp.getNext();

        o.plotFunction := pp.keywordPresent("-plotFunction");

        o.simpleMatch := pp.keywordPresent("-simpleMatch");

        pp.getKeyword("-minLength");
        o.minLength := pp.getNextLongReal();

        pp.getKeyword("-blurFactor");
        o.blurFactor := pp.getNextLongReal(0.0d0, 5.0d0);

        pp.getKeyword("-extraLength");
        o.extraLength := pp.getNextLongReal(0.0d0);

        pp.getKeyword("-maxShift");
        o.maxShift := pp.getNextLongReal();

        o.mp := PZMismatch.ParseOptions(pp);

        IF pp.keywordPresent("-maxCurveCands") THEN
          o.maxCurveCands := pp.getNextInt(1)
        ELSE
          o.maxCurveCands := LAST(NAT)
        END;

        IF pp.keywordPresent("-maxPairCands") THEN
          o.maxPairCands := pp.getNextInt(1, o.maxCurveCands)
        ELSE
          o.maxPairCands := LAST(NAT)
        END;

        IF pp.keywordPresent("-dontGuess") THEN
          o.dontGuess := TRUE;
        ELSE
          o.dontGuess := FALSE;
        END;

        o.perturbMatrix := pp.keywordPresent("-perturbMatrix");

        IF pp.keywordPresent("-dontOptimize") THEN
          o.maxEvals := 0
        ELSE
          pp.getKeyword("-maxEvals");
          o.maxEvals := pp.getNextInt();
        END;

        IF pp.keywordPresent("-displace") THEN
          o.displace := pp.getNextLongReal();
        ELSE
          o.displace := 0.0d0;
        END;

        IF pp.keywordPresent("-draw") THEN
          o.draw  := TRUE;
        ELSE
          o.draw := FALSE
        END;
        o.epsFormat := pp.keywordPresent("-epsFormat");
        IF pp.keywordPresent("-drawEvery") THEN
          o.drawEvery := pp.getNextInt(1, LAST(NAT))
        ELSE
          o.drawEvery := 1
        END;

        pp.finish();
      EXCEPT
      | ParseParams.Error =>
          Wr.PutText(stderr, "Usage: PZPZRefineCandsGeometry \\\n");
          Wr.PutText(stderr, "  -input NAME [ -maxCands NUMBER ] \\\n");
          Wr.PutText(stderr, "  [ -chainDir DIR ] -chainPrefix NAME \\\n");
          Wr.PutText(stderr, "  -band BAND -step NUMBER \\\n");
          Wr.PutText(stderr, "  [ -extension EXT ] \\\n");
          Wr.PutText(stderr, "  -output  NAME \\\n");
          Wr.PutText(stderr, "  [ -plotFunction ] \\\n");
          Wr.PutText(stderr, "  [ -simpleMatch ] \\\n");
          Wr.PutText(stderr, "  -maxShift NUM \\\n");
          Wr.PutText(stderr, "  " & PZMismatch.OptionsHelp & " \\\n");
          Wr.PutText(stderr, "  -minLength NUM \\\n");
          Wr.PutText(stderr, "  -blurFactor NUM -extraLength NUM \\\n");
          Wr.PutText(stderr, "  [ -maxCurveCands NUMBER ] \\\n");
          Wr.PutText(stderr, "  [ -maxPairCands NUMBER ] \\\n");
          Wr.PutText(stderr, "  [ -dontGuess ] [ -perturbMatrix ] \\\n");
          Wr.PutText(stderr, "  { -dontOptimize | -maxEvals  NUM } \\\n");
          Wr.PutText(stderr, "  [ -draw [ -epsFormat ] [ -drawEvery NUM ] ] \n");
          Process.Exit(1);
      END;
    END;
    RETURN o
  END  GetOptions;

PROCEDURE RefineCandidates(
    READONLY cand: PZCandidate.List;
    READONLY ch: ARRAY OF ChainData;
    READONLY o: Options;
    lambda: LONG;
  ) : REF PZCandidate.List =
  VAR
    c: REF PZCandidate.List;  (* Refined candidates. *)
    n: NAT := 0;         (* Refined candidate count. *)
    minAvgDist: REF LONGS := NIL;
    rCost: REF PZMatch.CostMatrix := NIL;
    plotName: TEXT;
  BEGIN
    WITH
      nCands = NUMBER(cand),
      nCurves = NUMBER(ch),
      length = NEW(REF LONGS, nCurves)^
    DO
      FOR k := 0 TO nCurves-1 DO
        IF ch[k].c # NIL THEN length[k] := ChainLength(ch[k].c^) END
      END;
      Wr.PutText(stderr, "candidates before refine loop = " & Fmt.Int(nCands) & "\n");
      c := NEW(REF PZCandidate.List, nCands);
      FOR i := 0 TO nCands-1 DO
        WITH 
          candTag = Fmt.Pad(Fmt.Int(i), 6, '0')
        DO
          Wr.PutText(stderr, "=== candidate " & candTag & " ========================\n");
          IF i = 0 AND o.plotFunction THEN
            plotName := o.output & "-" & candTag & "-fn"
          ELSE
            plotName := ""
          END;
          c[n] := RefineOneCandidate(
            cand[i], ch, o, lambda,
            plotName := plotName,
            drawName := o.output & "-" & candTag & "-dr",
            candCmt := CandComments(i, cand[i], o.step),
            (*WORK*)
            minAvgDist := minAvgDist,
            rCost := rCost
          );
          IF (i MOD 80 = 0) THEN Wr.PutText(stderr, "\n") END;
          IF c[n].mismatch <= 0.0d0 THEN
            Wr.PutText(stderr, "+");
            INC(n)
          ELSE
            Wr.PutText(stderr, "-");
          END;
        END;
      END;
      Wr.PutText(stderr, "\n");
      Wr.PutText(stderr, "candidates before pruning =      " & Fmt.Int(n) & "\n");
      PZCandidate.Sort(SUBARRAY(c^, 0, n), PZCandidate.PairMismatchBetter);
      PZCandidate.PruneCandsPerPair(c^, n, o.maxPairCands);
      Wr.PutText(stderr, "candidates after pair pruning =  " & Fmt.Int(n) & "\n");
      PZCandidate.Sort(SUBARRAY(c^, 0, n), PZCandidate.MismatchBetter);
      PZCandidate.PruneCandsPerCurve(c^, n, o.maxCurveCands);
      Wr.PutText(stderr, "candidates after curve pruning = " & Fmt.Int(n) & "\n");
      WITH ct = NEW(REF PZCandidate.List, n) DO
        ct^ := SUBARRAY(c^,0,n);
        RETURN ct
      END;
    END
  END RefineCandidates;

PROCEDURE ChainLength(READONLY c: PZLR3Chain.T): LONG =
  VAR s: LONG := 0.0d0;
  VAR j := LAST(c);
  BEGIN
    FOR i := 0 TO LAST(c) DO
      s := s + LR3.Dist(c[i], c[j]);
      j := i
    END;
    RETURN s
  END ChainLength;

PROCEDURE RefineOneCandidate(
    READONLY cOld: PZCandidate.T;
    READONLY ch: ARRAY OF ChainData;
    READONLY o: Options;
    lambda: LONG;
    plotName: TEXT;
    drawName: TEXT;
    candCmt: TEXT;
    (*WORK*)
    VAR minAvgDist: REF LONGS;
    VAR rCost: REF PZMatch.CostMatrix;
  ): PZCandidate.T =
  VAR
    mt: REF PZMatch.T;
    mismatch: LONG;
    candLength: LONG;
    matchedLength: LONG;
    mat: LR4x4.T;
    aMap:  REF PZLR3Chain.T;
    extra: ARRAY [0..1] OF INT;
  BEGIN
    (* Compute the expansion needed, in samples: *)
    FOR j := 0 TO 1 DO
      extra[j] := CEILING(o.extraLength / o.step)
    END;

    WITH
      cExp = PZCandidate.Expand(cOld, extra, extra),

      cvxa = cOld.seg[0].cvx,
      ca = PZLR3Chain.ExtractSegment(ch[cvxa].c^, cExp.seg[0])^,

      cvxb = cOld.seg[1].cvx,
      cb = PZLR3Chain.ExtractSegment(ch[cvxb].c^, cExp.seg[1])^
    DO
      <* ASSERT NUMBER(ch[cvxa].c^) = cOld.seg[0].tot *>
      <* ASSERT NUMBER(ch[cvxb].c^) = cOld.seg[1].tot *>
      PZCandidate.Print(stderr, cExp);
      ComputeMatrix(
        a := ca,
        b := cb,
        o := o,
        step := o.step,
        minLength := o.minLength - o.blurFactor * 2.0d0 * lambda,
        plotName := plotName,
        drawName := drawName,
        candCmt := candCmt,
        (*OUT*)
        mat := mat,
        mt := mt,
        mismatch := mismatch,
        length := candLength,
        matchedLength := matchedLength,
        aMap:= aMap,
        (*WORK*)
        minAvgDist := minAvgDist,
        rCost := rCost
      );
      Wr.PutText(stderr, "\n");
      WITH cNew = PZCandidate.TrimAndPack(
          cExp, mt^,
          mismatch := mismatch,
          length := candLength,
          matchedLength := matchedLength
        )
      DO
        PZCandidate.Print(stderr, cNew);
        RETURN cNew
      END
    END
  END RefineOneCandidate;

PROCEDURE ComputeMatrix(
    READONLY a, b: PZLR3Chain.T;   (* Contour segments to match. *)
    READONLY o: Options;
    step: LONG;                (* Mean sampling step. *)
    minLength: LONG;           (* Minimum length of matching. *)
    drawName: TEXT;                (* File name prefix for drawings. *)
    plotName: TEXT;                (* File name prefix for goal function plots. *)
    candCmt: TEXT;                 (* Candidate description for plot caption. *)
    (*OUT*)
    VAR mat: LR4x4.T;
    VAR mt: REF PZMatch.T;
    VAR mismatch: LONG;
    VAR length: LONG;
    VAR matchedLength: LONG;
    VAR aMap: REF PZLR3Chain.T;
    (*WORK*)
    VAR minAvgDist: REF LONGS;
    rCost: REF PZMatch.CostMatrix;
  ) =

  VAR f, f1: Value;
      nEvals: NAT;
  BEGIN
    WITH
      ma = NUMBER(a),
      mb = NUMBER(b),
      Ra = Dist(a[0], a[ma-1])/2.0d0,
      Rb = Dist(b[0], b[mb-1])/2.0d0,
      R = (Ra + Rb)/2.0d0,
      scaleY = Math.sqrt(1.0d0/3.0d0) * R,
      scaleX = scaleY / 4.0d0,
      minChainEdges = MAX(0, FLOOR((2.0d0 * minLength)/step))
    DO

      PROCEDURE GoalFunction(READONLY mat: LR4x4.T): LONG =
        (*
          Computes the image "aMap" of chain "a" by "mat", the best
          matching "mt" of "aMap" with "b", and returns the mismatch
          if negative.
          
          Otherwise the mismatch must be zero; in that case, returns
          the distance squared between the midpoint of "a" and 
          the nearest point of "b".
        *)
        BEGIN
          aMap := PZLR3Chain.Map(a, mat);
          ComputeMatch(aMap^, b, o,
            minChainEdges := minChainEdges,
            step := step,
            mt := mt,
            mismatch := mismatch,
            length := length,
            matchedLength := matchedLength,
            minAvgDist := minAvgDist,
            rCost := rCost
          );
          RETURN mismatch
        END GoalFunction;

      BEGIN
        (* Compute an initial guess for the placement matrix: *)
        mat := ComputeInitialMatrix(a, b, o);

        (* Compute initial mapped chain, matching, and mismatch: *)
        f := GoalFunction(mat);
        <* ASSERT f = mismatch *>

        (* Plot initial placement and matching: *)
        IF o.draw THEN
          DrawPlacement(drawName & "-i", o, candCmt,
            aMap^, b, mt^, nEvals,
            mismatch := mismatch,
            length := length,
            matchedLength := matchedLength
          )
        END;

        (* Optimize matrix *)
        PZOptMatrix.Optimize(
          func := GoalFunction,
          scaleX := scaleX, scaleY := scaleY,
          maxEvals := o.maxEvals,
          nEvals := nEvals,
          m := mat,
          f := f,
          plotName := plotName
        );

        (* Compute final mapped chainh, matching, and mismatch: *)
        f1 := GoalFunction(mat);
        <* ASSERT f1 = f *>
        <* ASSERT f = mismatch *>

        (* Plot optimum placement and matching: *)
        IF o.draw THEN
          DrawPlacement(drawName & "-f", o, candCmt,
            aMap^, b, mt^, nEvals,
            mismatch := mismatch,
            length := length,
            matchedLength := matchedLength
          )
        END;
      END
    END
  END ComputeMatrix;

PROCEDURE ComputeInitialMatrix(
    READONLY a, b: PZLR3Chain.T;
    READONLY o: Options;
  ): LR4x4.T =
  VAR mat: LR4x4.T;
  BEGIN
    IF o.dontGuess THEN
      mat := Identity();
    ELSE
      mat := PZLR3Chain.AlignmentMatrix(a, b)
    END;
    IF o.perturbMatrix THEN
      mat := PerturbMatrix (mat, dTheta := 0.1d0, dX := 1.0d0, dY := 1.0d0)
    END;
    RETURN mat
  END ComputeInitialMatrix;

PROCEDURE PerturbMatrix(VAR mat: LR4x4.T; dTheta, dX, dY: LONG): LR4x4.T =
  VAR n: LR4x4.T := LR4x4.Identity();
  BEGIN
    WITH
      cos = Math.cos(dTheta),
      sin = Math.sin(dTheta)
    DO
      <* ASSERT cos # 0.0d0 OR sin # 0.0d0 *>
      n[1,1] := cos;
      n[1,2] := sin;
      n[2,1] := -sin;
      n[2,2] := cos;
      n[0,1] := dX;
      n[0,2] := dY;
    END;
    RETURN LR4x4.Mul(mat, n)
  END PerturbMatrix;

PROCEDURE ComputeMatch(
    READONLY a, b: PZLR3Chain.T;
    READONLY o: Options;
    minChainEdges: NAT;
    step : LONG;
    (*OUT*)
    VAR mt: REF PZMatch.T;
    VAR mismatch: LONG;
    VAR length: LONG;
    VAR matchedLength: LONG;
    (*WORK*)
    VAR minAvgDist: REF LONGS;
    rCost: REF PZMatch.CostMatrix;
  ) =
  VAR avgDist: LONG;
  BEGIN
    WITH
      NA = NUMBER(a),
      NB = NUMBER(b),
      ctr = PickCenter(a, b),
      mtOld = PZMatch.T{ctr}
    DO
      IF minChainEdges <= (NA-1)+(NB-1) THEN
        IF o.simpleMatch THEN
          PZLR3Chain.FindBestMatch(
            a, b,
            maxDistSqr := o.mp.maxDistSqr,
            removeUnmatchedEnds := TRUE, 
            (*OUT*)
            mt := mt,
            avgDist := avgDist,
            length := length,
            matchedLength := matchedLength,
            (*WORK*)
            rCost := rCost
          );
          mismatch := (avgDist*avgDist - o.mp.critDistSqr)*length
        ELSE
          IF minAvgDist = NIL OR NUMBER(minAvgDist^) < NA + NB - 1 THEN
            minAvgDist := NEW(REF LONGS, NA + NB - 1)
          END;
          PZLR3Chain.RefineMatch(
            a, b,
            step := step,
            mtOld := mtOld,
            maxAdjust := CEILING(o.maxShift / o.step),
            critDistSqr := o.mp.critDistSqr,
            maxDistSqr := o.mp.maxDistSqr,
            (*OUT*)
            mt := mt,
            mismatch := mismatch,
            length := length,
            matchedLength := matchedLength,
            (*WORK*)
            minAvgDist := SUBARRAY(minAvgDist^, 0, NA+NB-1),
            rCost := rCost
          );
        END;
        (* If the matching has zero length, provide a useful "mismatch": *)
        IF mt = NIL OR NUMBER(mt^) < 2 OR length <= 0.0d0 THEN 
          <* ASSERT mismatch = 0.0d0 *>
          WITH
            ac = (NA-1) DIV 2,
            bc = IndexNearestPoint(b, a[ac])
          DO
            mismatch := LR3.DistSqr(a[ac], b[bc]);
          END
        END
      ELSE
        mismatch := LAST(LONG);
        length := 0.0d0;
        matchedLength := 0.0d0;
        mt := NEW(REF PZMatch.T, 1);
        mt[0] := ctr
      END
    END;
  END ComputeMatch;
  
PROCEDURE PickCenter(READONLY a, b: PZLR3Chain.T): PZMatch.Pair =
  BEGIN
    WITH
      ai = IndexNearestPoint(a, b[0]),
      af = IndexNearestPoint(a, b[LAST(b)]),
      bi = IndexNearestPoint(b, a[0]),
      bf = IndexNearestPoint(b, a[LAST(a)])
    DO
      RETURN PZMatch.Pair{(ai + af) DIV 2, (bi + bf) DIV 2}
    END
  END PickCenter;

PROCEDURE IndexNearestPoint(READONLY b: PZLR3Chain.T; p: LR3.T ): NAT =
  VAR index: NAT;
      mindist:  LONG;
  BEGIN
     WITH
       m = NUMBER(b),
       mHalf = (m-1) DIV 2
     DO
       index := 0; mindist := Dist(b[0],p);
       FOR i:= mHalf TO 1 BY -1  DO
         WITH d = Dist(b[i],p) DO
           IF d <  mindist THEN index := i; mindist := d END;
         END
       END;
       FOR i:= mHalf+1 TO m-1 DO
         WITH d = Dist(b[i],p) DO
           IF d <  mindist THEN index := i; mindist := d END;
         END
       END;
     END;
     RETURN index
 END IndexNearestPoint;

PROCEDURE DrawPlacement(
    fileName: TEXT;               (* File name without extension. *)
    READONLY o: Options;
    candCmt: TEXT;                (* Candidate description for caption. *)
    READONLY a, b: PZLR3Chain.T;  (* Partial contours, reversed and mapped. *)
    READONLY mt: PZMatch.T;       (* The match, relative to "a" and "b". *)
    nEvals: NAT;                  (* Goal function evaluations done. *)
    mismatch: LONG;               (* Mismatch of "mt". *)
    length: LONG;                 (* Total length of "a" and "b" *)
    matchedLength: LONG;          (* Length actually matched. *)
  ) =
  BEGIN
    WITH
      ma = NUMBER(a),
      mb = NUMBER(b),
      w = LR3.T{1.0d0, 0.0d0, 0.0d0},
      p = a[0],
      q = a[LAST(a)],
      u = PZGeo.SegDir(p, q),
      mat = PZGeo.Rotation(u, w),
      cand = PZCandidate.T{
        seg := PZSegment.Pair{
          PZSegment.T{cvx := 0, tot := ma, ini := 0, ns := ma, rev := FALSE},
          PZSegment.T{cvx := 0, tot := mb, ini := 0, ns := mb, rev := FALSE}
        },
        mismatch := mismatch,
        length := length,
        matchedLength := matchedLength,
        pm := PZMatch.Pack(mt)
      },
      fd = o.displace/2.0d0,
      am = DisplaceChain(PZLR3Chain.Map(a, mat)^, LR3.T{0.0d0, +fd, 0.0d0})^,
      bm = DisplaceChain(PZLR3Chain.Map(b, mat)^, LR3.T{0.0d0, -fd, 0.0d0})^,
      caption = candCmt & MatchComments(
        nEvals, mat, mt, mismatch, length, matchedLength
      )
    DO
      PZFullPlot.Candidate(
        fileName,
        cand,
        c0 := am, c1 := bm,
        step := o.drawEvery,
        whole := FALSE,
        closed := FALSE,
        dots := FALSE,
        pairing := TRUE,
        pointers := FALSE,
        labels := FALSE,
        axes := FALSE,
        epsFormat := o.epsFormat,
        caption := caption
      );
    END;
    Wr.PutText( stderr, ".");
  END DrawPlacement;

PROCEDURE DisplaceChain(READONLY a: PZLR3Chain.T; READONLY d: LR3.T): REF PZLR3Chain.T =
  BEGIN
    WITH
      r = NEW(REF PZLR3Chain.T, NUMBER(a)), ad = r^
    DO
      FOR i := 0 TO LAST(a) DO ad[i] := LR3.Add(a[i], d) END;
      RETURN r
    END
  END DisplaceChain;

PROCEDURE CandComments(
    iCand: NAT;
    READONLY cand: PZCandidate.T;
    step: LONG;
  ): TEXT =
  BEGIN
    RETURN
      "Candidate " & Fmt.Pad(Fmt.Int(iCand), 6, '0') &
      "  step = " & Fmt.LongReal(step, prec := 6) &
      "\n" &
      SegmentComments(cand.seg[0]) &
      SegmentComments(cand.seg[1])
  END CandComments;

<* UNUSED *>
PROCEDURE FI(x: INT; w: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.Int(x), w, '0')
  END FI;

PROCEDURE SegmentComments(READONLY s: PZSegment.T): TEXT =
  BEGIN
    WITH
      wr = NEW(TextWr.T).init()
    DO
      Wr.PutText( wr, "  chain " & Fmt.Pad(Fmt.Int(s.cvx), 4, '0'));
      Wr.PutText( wr, " (total " & Fmt.Pad(Fmt.Int(s.tot), 5) & " samples)");
      Wr.PutText( wr, " " & Fmt.Pad(Fmt.Int(s.ns), 5) & " samples");
      Wr.PutText( wr, " from sample " & Fmt.Pad(Fmt.Int(s.ini), 5));
      IF s.rev THEN Wr.PutText( wr, " (reversed)") END;
      Wr.PutText( wr, "\n");
      RETURN TextWr.ToText(wr)
    END
  END SegmentComments;

PROCEDURE MatchComments(
    nEvals: NAT;
    READONLY mat: LR4x4.T;
    READONLY mt: PZMatch.T;
    mismatch: LONG;
    length: LONG;
    matchedLength: LONG;
 ): TEXT =
  BEGIN
    WITH
      wr = NEW(TextWr.T).init(),
      nm = NUMBER(mt)
    DO
      Wr.PutText(wr, "  matched ");
      Wr.PutText(wr, FmtMatchedRange(mt[0][0], mt[nm-1][0]));
      Wr.PutText(wr, " to ");
      Wr.PutText(wr, FmtMatchedRange(mt[0][1], mt[nm-1][1]));
      Wr.PutText(wr, "  nEvals = " & Fmt.Int(nEvals) & "\n");
      Wr.PutText(wr, "  (cos,sin) = " & PairToText(mat[1,1], mat[1,2]) & "\n");
      Wr.PutText(wr, "  (dX,dY) = " & PairToText(mat[0,1], mat[0,2]) & "\n");
      Wr.PutText(wr, "  mismatch = " & Fmt.LongReal(mismatch, prec := 6));
      Wr.PutText(wr, "  length = " & Fmt.LongReal(length, prec := 6));
      Wr.PutText(wr, "  matchedLength = " & Fmt.LongReal(matchedLength, prec := 6) & "\n");
      RETURN TextWr.ToText(wr)
    END
  END MatchComments;

PROCEDURE RunComments(READONLY o: Options): TEXT =
  BEGIN
    WITH
      wr = NEW(TextWr.T).init()
    DO
      Wr.PutText(wr, "  input = " & o.input & "\n");
      Wr.PutText(wr, "  maxCands = " & Fmt.Int(o.maxCands) & "\n");
      Wr.PutText(wr, "  chainDir = " & o.chainDir);
      Wr.PutText(wr, "  chainPrefix =  "  & o.chainPrefix);
      Wr.PutText(wr, "  band = "  & Fmt.Int(o.band) & "\n");
      Wr.PutText(wr, "  output = " & o.output & "\n");
      Wr.PutText(wr, "  maxShift = "  & Fmt.LongReal(o.maxShift, prec := 6) &  "\n");
      Wr.PutText(wr, "  maxDist = "  & Fmt.LongReal(sqrt(o.mp.maxDistSqr), prec := 6) &  "\n");
      Wr.PutText(wr, "  critDist = "  & Fmt.LongReal(sqrt(o.mp.critDistSqr), prec := 6));
      Wr.PutText(wr, "  minLength = "  & Fmt.LongReal(o.minLength, prec := 6) &  "\n");
      Wr.PutText(wr, "  blurFactor = " & Fmt.LongReal(o.blurFactor, prec := 6));
      Wr.PutText(wr, "  extraLength = " & Fmt.LongReal(o.extraLength, prec := 6) & "\n");
      RETURN TextWr.ToText(wr)
    END
  END RunComments;

PROCEDURE FmtMatchedRange(iniMatch, finMatch: NAT): TEXT =
  BEGIN
    WITH wr = NEW(TextWr.T).init() DO
      Wr.PutText(wr, " [" & Fmt.Int(iniMatch));
      Wr.PutText(wr, ".." & Fmt.Int(finMatch) & "]");
      RETURN TextWr.ToText(wr)
    END
  END FmtMatchedRange;

PROCEDURE PairToText(x, y: LONG): TEXT =
  BEGIN
    RETURN "(" & Fmt.LongReal(x, prec := 6) & ", " & Fmt.LongReal(y, prec := 6) & ")"
  END PairToText;

BEGIN
  Main()
END PZRefineCandsGeometry.

(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
