INTERFACE PZMatch;

(* Chain distance by dynamic programming. *)
(* Last edited on 1999-11-15 12:39:47 by hcgl *)

IMPORT Rd, Wr;
IMPORT PZSegment;
FROM PZTypes IMPORT LONG, LONGS, NAT, BOOL;

TYPE 
  Pair = ARRAY [0..1] OF NAT;
  T = ARRAY OF Pair;
  (*
    A "PZMatch.T" is a `monotone matching' between two arbitrary sequences
    "a[0..NA-1]" and "b[0..NB-1]", defined as a list of pairs
    "mt[0..m-1] = (r[0],s[0]),.. (r[m-1],s[m-1])" such that 
    
    | r[0] >= 0, s[0] >= 0;
    | 
    | r[m-1] <= NA-1, s[m-1] <= NB-1;
    | 
    | r[k+1] = r[k]  OR  r[k+1] = r[k]+1,
    | s[k+1] = s[k]  OR  s[k+1] = s[k]+1,
    |
    | mt[k+1] # mt[k]
    
    for all "k".  By definition, element "a[r[k]] = a[mt[k][0]]" is
    paired with sample "b[s[k]] = b[mt[k][1]]", for all "k".

    We say that a pair "(r1,s1)" `precedes' another pair "(r2,s2)" if
    "r1 <= r2" and "s1 <= s2", and at least one of these inequalities
    is strict.  In that case we also say that "(r2,s2)" `follows'
    "(r1,s1)".  Note that, in any monotone matching "mt", "mt[i]"
    precedes "mt[j]" iff "i < j".
    
    Two consecutive pairs "mt[k]=(r[k],s[k])" and
    "mt[k+1]=(r[k+1],s[k+1])" are said to be a `step' of the match.
    The step is said to be `full' if both "r" and "s" increase by 1,
    and `half' if only one of them increases.  A `perfect matching' is
    one that uses only full steps.
    
    The matching is said to be `complete' if "mt[0] = (0,0)" and
    "mt[m-1] = (NA-1,NB-1)"; and `incomplete' otherwise.
    
  *)
  
(* MINIMUM-COST MATCHINGS *)

(*
  The procedures in this section take as argument an arbitrary `step
  cost function' "stepCost" (which may return negative values).

  The `total cost' of a matching "mt" is the sum of "stepCost(s[k])"
  over all steps "S[k] = mt[k]->mt[k+1]" of "mt".
  
  The `step count' of a match is defined as the number of
  half-steps plus twice the number of full steps.  It is also the
  number of samples matched in the two ckains, minus two; thus any
  complete matching has step count "NA+NB-2".
  
  The optimization procedures also use internally a "CostMatrix" "rCost"
  with at least "NA" rowns and "NB" columns, which may be preallocated
  and given by the client.  The procedures will (re)allocate said
  matrix if it is NIL or too small. 
  
*)

TYPE
  StepFunction = PROCEDURE (READONLY p1, p2: Pair): LONG;

TYPE
  CostMatrix = ARRAY OF LONGS;  (* Work array used by DP algorithm *)

PROCEDURE FindBest(
    NA, NB: NAT; 
    stepCostFn: StepFunction;
    VAR (*OUT*) mt: REF T;
    VAR (*OUT*) totCost: LONG;
    VAR (*WORK*) rCost: REF CostMatrix;
  );
  (*
    Returns in "mt" a complete monotone matching between 
    two strings with "NA" and "NB" elements, with mimimum
    total cost (which is returned in "totCost". *)

PROCEDURE Refine(
    NA, NB: NAT;
    stepCostFn: StepFunction;
    READONLY mtOld: T;
    maxAdjust: NAT;
    VAR (*OUT*) mt: REF T;
    VAR (*OUT*) totCost: LONG;
    VAR (*OUT*) minTotCost: LONGS;
    VAR (*WORK*) rCost: REF CostMatrix;
  );
  (*
    Computes a partial monotone matching, between two arbitrary
    sequences with "NA" and "NB" elements, that is not too far from "mtOld"
    and has minimum total cost.
    
    The `input matching' "mtOld" is an arbitrary (but non-empty)
    matching between the two sequences, which is to be improved by the
    procedure.
    
    On output, "mt" will point to the matching of minimum total cost
    found by the procedure, within certain constraints (or to NIL if
    the constraints cannot be satisfied). The total cost of "mt"
    will be returned in "totCost".

    Let's define the `initial pairs' as consisting of all pairs of the
    input matching "mtOld", plus all possible index pairs that precede
    its first pair, plus all possible index pairs that follow its last
    pair.  One constraint on the refined matching "mt" is that is
    that every pair "mt[k]" must lie within "maxAdjust" distance
    from some initial pair, in the Manhattan metric.  In other words,
    the `valid pairs' for the refined matching are the union of all
    sets of the form "{ (r+dr,s+ds) : |dr| <= maxAdjust, |ds| <=
    maxAdjust }" where "(r,s)" is an initial pair.

    A second constraint is that the refined matching "mt" will
    contain exactly one pair "(ro,so)" such that "ro+so = rc+sc",
    where "(rc,sc)" is the middle pair of the input matching "mtOld".
    (This constraint is meant to erclude partial matchings that cover
    only short initial or final segments of the two sequences.)
    
    Thus, the returned match will be one that:
    
      (1) contains only valid pairs, as defined above;
      
      (2) begins with some pair "(ri,si)" with "ri+si <= rc+sc";
      
      (3) ends with some pair "(rf,sf)" with "rf+sf >= rc+sc";
      
      (4) has minimum total cost, among all matchings that satisfy (1)-(3).
      
    Note that there is always a matching that satisfies (1)-(3), namely the 
    matching "mtOld" itself (or any piece of it that includes
    its central pair).
      
    The procedure also returns in "minTotCost" a table such that
    "minTotCost[k]" is the minimum total cost (NOT the mismatch) of
    any matching with "k" steps that satisfies the above constraints,
    for each "k" in "[0..NA+NB-2]".
  *)

(* UTILITY FUNCTIONS *)

PROCEDURE Trim(READONLY mt: T; start, count: NAT): REF T;
  (*
    Returns the segment "mt[start..start+count-1]" of matching "mt". *)

PROCEDURE Adjust(
    READONLY mt: T; 
    READONLY sOld: PZSegment.Pair;
    READONLY sNew: PZSegment.Pair;
  ): REF T;
  (*
    Given a matching between segments "sOld[0]" and "sOld[1]", returns
    the equivalent matching between "sNew[0]" and "sNew[1]".  This
    entails mapping the relative sample indices in "mt" so that they
    refer to the same samples, and possibly trimming pairs that fall
    outside the new segments.
    
    The result is NIL if none of the old pairs are contained in the
    new segments, or the intersection of the restriction of "mt"
    to the new segments is not a single, continuous matching. *)

PROCEDURE RemoveDanglingEnds(READONLY mt: T): REF T;
  (*
    Returns a new matching obtained by removing
    the ``dangling ends'' of "mt", i.e. all initial and final
    half-steps of "mt" where one of the ``legs'' remains stuck at
    the initial or final point of the corresponding chain. *)

PROCEDURE MostPerfect(n0, n1: NAT): REF T;
PROCEDURE DoMostPerfect(n0, n1: NAT; VAR mt: T);
  (*
    Generates a maximally perfect matching between two segments
    with "n0" and "n1" samples, respectively. The corrresponding
    path is a Bresenham line between "(0,0)" and "(n0,n1)"". 
    
    DoMostPerfect" is like "MostPerfect" but stores the
    matching in the client-provided array "mt", which should
    have "MAX(n0,n1)" elements exactly. *)
    
PROCEDURE AbsPairClip(p: Pair; READONLY s: PZSegment.Pair): Pair;
  (*
    Given a pair "p" of sample indices, relative to segments "s[0]"
    and "s[1]", returns the corresponding indices in the whole curves. *)

PROCEDURE RelPairClip(p: Pair; READONLY s: PZSegment.Pair): Pair;
  (*
    Given a pair "p" of sample indices in two curves, 
    returns the corresponding sample indices relative to segments "s[0]"
    and "s[1]".  Each returned index "p[j]" will be "LAST(NAT)" 
    if the corresponding sample lies outside the segment "s[j]. *)

TYPE
  StepPredicate = PROCEDURE (READONLY p1, p2: Pair): BOOL;
  StepMeasure = PROCEDURE (READONLY p1, p2: Pair): LONG;

PROCEDURE RemoveUnmatchedEnds(
    READONLY mt: T; 
    isMatched: StepPredicate;
  ): REF T;
  (*
    Removes all initial and final steps "mt[k]->mt[k+1]" of "mt" for which
    "isMatched(mt[k],mt[k+1])" returns FALSE.  *)
    
PROCEDURE Length(
    READONLY mt: T; 
    lengthFn: StepMeasure;
  ): LONG; 
  (*
    Returns the sum of of "lengthFn(S)" for all steps of the matching "mt". *)

PROCEDURE Sum(
    READONLY mt: T; 
    stepFn: StepFunction;
  ): LONG; 
  (*
    Returns the sum of "stepFn(S)" over the matching "mt". *)

PROCEDURE Integral(
    READONLY mt: T; 
    lengthFn: StepMeasure;
    stepFn: StepFunction;
  ): LONG; 
  (*
    Returns the integral of "stepFn(S)" over the matching "mt",
    with respect to step length as returned by "lengthFn(S)". *)

PROCEDURE MatchedStepCount(
    READONLY mt: T; 
    isMatched: StepPredicate;
  ): NAT;
  (*
    Returns the step count of "mt" (number of half-steps plus twice
    the number of full steps), considering only steps "S"
    for which "isMatched(S)" is TRUE. *)
 
PROCEDURE MatchedLength(
    READONLY mt: T; 
    lengthFn: StepMeasure;
    isMatched: StepPredicate;
  ): LONG;
  (*
    Returns the sum of "lengthFn(S)" for all steps "S" in "mt" 
    for which "isMatched(S)" is true. *)
    
PROCEDURE Overlap(
    READONLY a, b: T;
    READONLY aSeg, bSeg: PZSegment.Pair;
    maxAdjust: NAT;
  ): NAT;
  (*
    Compares two matchings "a" and "b", which are relative to segment
    pairs "aSeg" and "bSeg", respectively.  Returns the number of
    half-steps of "a" that lie at most "maxAdjust" from some half-step of
    "b" with same position.  Note that "Overlap(a,b,...) = Overlap(b,a,...)". *)

(* PACKED MATCHINGS *)

TYPE
  PackedT = RECORD  (* A compact description of a matching *)
      np: NAT;      (* Number of pairs (> 0). *)
      ini:  Pair;   (* Initial pair *) 
      fin: Pair;    (* Final pair *)
      steps: TEXT;  (* Encoded steps *)
    END;
    (*
      Code for encoded steps '/' advance [0], '\\' advance [1], '|' both. *)
    
PROCEDURE Pack(READONLY mt: T): REF PackedT;
  (*
    Returns a packed encoding of "mt". *)
    
PROCEDURE PackPerfect(nSamples: NAT): REF PackedT;
  (*
    Returns a packed perfect matching between two
    segments, both with "nSamples" segments. *)
    
PROCEDURE Unpack(READONLY pm: PackedT): REF T;
PROCEDURE DoUnpack(READONLY pm: PackedT; VAR mt: T);
  (*
    Expands a packed representation into an explicit matching.
    "DoUnpack" is like "Unpack" but returns the result in 
    the client-provided array "mt". *)

PROCEDURE WritePacked(wr: Wr.T; READONLY pm: PackedT);
  (*
    Writes the matching "pm" in compact form to writer "wr". *)
    
PROCEDURE ReadPacked(rd: Rd.T): REF PackedT;
  (*
    Reads from "rd" a packed matching, in the format written by "Write". *)

PROCEDURE AdjustPacked(rpm: REF PackedT;
    READONLY oldSeg: PZSegment.Pair;
    READONLY newSeg: PZSegment.Pair;
  ): REF PackedT;
  (*
    Given a packed matching "rpm^" between two segments
    "oldSeg[0]" and "oldSeg[1]", returns a new packed matching 
    that describes the same pairing between the segments
    "newSeg[0]" and "newSeg[1]".  If the new segments are
    not contained in the original ones, the matching may be
    trimmed. 
    
    The result is NIL if "rpm" is NIL or if the intersection of the
    new segments and the old pairings is not a single interval. *)

END PZMatch.

(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
