MODULE PZMatch;

(* Last edited on 1999-11-19 10:25:14 by hcgl *)

IMPORT Rd, Wr, TextWr, Thread, Fmt, Text, FPut, FGet;
IMPORT PZSegment;
FROM PZSegment IMPORT AbsIndexClip, RelIndexClip;
FROM PZTypes IMPORT LONG, LONGS, INT, INTS, NAT, BOOL;

FROM Stdio IMPORT stderr; (*DEBUG*)
<* FATAL Thread.Alerted, Wr.Failure *>

TYPE
  Direction = [-1..+1]; (* -1 = prefix, +1 = suffix *)

PROCEDURE AllocCostMatrix(VAR rCost: REF CostMatrix; NA, NB: NAT) =
  (*
    Ensures that "rCost^" exists and has at least "NA" rows and "NB"
    columns (but perhaps more). *)
  BEGIN
    IF rCost = NIL
    OR NUMBER(rCost^) < NA
    OR NUMBER(rCost^[0]) < NB THEN
      rCost := NEW(REF CostMatrix, NA, NB)
    END;
  END AllocCostMatrix;

PROCEDURE FindBest(
    NA, NB: NAT;
    stepCostFn: StepFunction;
    VAR (*OUT*) match: REF T;
    VAR (*OUT*) totCost: LONG;
    VAR (*WORK*) rCost: REF CostMatrix;
  ) =
  VAR k: NAT;
  BEGIN
    <* ASSERT NA > 0 AND NB > 0 *>
    AllocCostMatrix(rCost, NA, NB);
    WITH
      c = rCost^,
      np = NA+NB-1,
      rp = NEW(REF T, np),
      p = rp^
    DO
      (* Fill the cost matrix: *)
      (*
        We set "c[ix,iy]" with the cost of the cheapest partial
        matching that begins with "(0,0)" and ends with "(ix,iy)"
      *)
      c[0,0] := 0.0d0;
      FOR ix := 1 TO NA-1 DO
        c[ix,0] := c[ix-1,0] + stepCostFn(Pair{ix-1,0}, Pair{ix,0});
      END;
      FOR iy := 1 TO NB-1 DO
        c[0,iy] := c[0,iy-1] + stepCostFn(Pair{0,iy-1}, Pair{0,iy});
      END;
      FOR ix := 1 TO NA-1 DO
        FOR iy := 1 TO NB-1 DO
          WITH
            v1 = c[ix-1,iy] + stepCostFn(Pair{ix-1,iy}, Pair{ix,iy}),
            v2 = c[ix,iy-1] + stepCostFn(Pair{ix,iy-1}, Pair{ix,iy}),
            v3 = c[ix-1,iy-1] + stepCostFn(Pair{ix-1,iy-1}, Pair{ix,iy})
          DO
            c[ix,iy] := MIN(MIN(v1,v2),v3)
          END
        END
      END;

      (* Compute the optimum matching: *)
      (*
        We trace back from "c[NA-1,NB-1]" to "c[0,0]", reconstructing the
        choice that resulted in the best matching at each step:
      *)
      k := np-1;
      p[k] := Pair{NA-1,NB-1};
      WHILE p[k][0] > 0 OR p[k][1] > 0 DO
        WITH
          ix = p[k][0],
          iy = p[k][1]
        DO
          IF ix = 0 THEN
            p[k-1] := Pair{ix,iy-1}
          ELSIF iy = 0 THEN
            p[k-1] := Pair{ix-1,iy}
          ELSE
            WITH
              v0 = c[ix,iy],
              v1 = c[ix-1,iy] + stepCostFn(Pair{ix-1,iy}, Pair{ix,iy}),
              v2 = c[ix,iy-1] + stepCostFn(Pair{ix,iy-1}, Pair{ix,iy}),
              v3 = c[ix-1,iy-1] + stepCostFn(Pair{ix-1,iy-1}, Pair{ix,iy})
            DO
              IF v0 = v3 THEN
                p[k-1] := Pair{ix-1,iy-1}
              ELSIF v0 = v1 THEN
                p[k-1] := Pair{ix-1,iy}
              ELSIF v0 = v2 THEN
                p[k-1] := Pair{ix,iy-1}
              ELSE
                <* ASSERT FALSE *>
              END
            END
          END
        END;
        DEC(k);
      END; (* while *)
      totCost := c[NA-1,NB-1] - c[0,0];
      match := Trim(p, k, np-k)
    END;
  END FindBest;

PROCEDURE Refine(
    NA, NB: NAT;
    stepCostFn: StepFunction;
    READONLY mtOld: T;
    maxAdjust: NAT;
    VAR (*OUT*) mt: REF T;
    VAR (*OUT*) totCost: LONG;
    VAR (*OUT*) minTotCost: LONGS;
    VAR (*WORK*) rCost: REF CostMatrix;
  ) =
  VAR mBest: NAT;
  BEGIN
    (*
      This code uses the following concepts.  The `position' of an
      index pair "(r,s)" is the sum "r+s", and its `alignment' is the
      difference "r-s".

      Obviously the the pair positions are strictly increasing as one
      moves along a match; while their alignments may vary in either
      direction, although they should remain roughly constant for
      matchings that are approximately 1-1.

      Note that the position "p" and alignment "a" of a pair have the
      same parity, and that "r = (p+a) DIV 2, s = (p-a) DIV 2".  Also
      note that the alignments of a set of valid pairs with a given
      position "p" comprise a single interval of integers with same
      parity as "p". This observation allows us to represent
      the set of valid pairs by two vectors "minAlign[p]"
      and "maxAlign[p]", where "p" ranges from "minPos = 0+0"
      to "maxPos = (NA-1)+(NB-1)".
    *)

    <* ASSERT NA > 0 AND NB > 0 *>
    <* ASSERT NUMBER(mtOld) > 0 *>

    AllocCostMatrix(rCost, NA, NB);

    (* Initialize the "minTotCost" table of min cost paths by step count. *)
    <* ASSERT LAST(minTotCost) = NA+NB-2 *>
    FOR k := 0 TO LAST(minTotCost) DO minTotCost[k] := LAST(LONG) END;

    WITH
      (* The range of the `position' coordinate of valid pairs. *)
      minPos = 0,
      maxPos = (NA-1)+(NB-1),

      (* Range of the `alignment' coordinate of valid pairs, for each `position'. *)
      minAlign = NEW(REF INTS, maxPos + 1)^,
      maxAlign = NEW(REF INTS, maxPos + 1)^,

      (* The central pair of the input matching: *)
      mOld = NUMBER(mtOld),
      mid = mtOld[mOld DIV 2],
      midPos = mid[0] + mid[1],

      (* The current best matching: *)
      maxPairs = NA + NB - 1,
      mtBest = NEW(REF T, maxPairs)^,

      (* Work areas for "RefineAux": *)
      c = rCost^,
      cMin = NEW(REF LONGS, maxPos + 1)^
    DO
      <* ASSERT minPos = 0 *>
      ComputeAlignmentRanges(NA, NB, mtOld, maxAdjust, minAlign, maxAlign);
      totCost := LAST(LONG);
      mBest := 0;
      FOR dPos := 0 TO 1 DO
        (* Wr.PutText(stderr,"dPos: "  &  Fmt.Int(dPos) & "\n"); *)
        WITH ctrPos = midPos + dPos DO
          (* DebugIntPair("NA,NB:", NA,NB); *)
          FOR ctrAlign := minAlign[ctrPos] TO maxAlign[ctrPos] BY 2 DO
            <* ASSERT ctrAlign MOD 2 = ctrPos MOD 2 *>
            WITH
              rc = (ctrPos+ctrAlign) DIV 2,
              sc = (ctrPos-ctrAlign) DIV 2
            DO
            (*  DebugIntPair("rc, sc:",rc, sc); *)
              <* ASSERT rc >= 0 AND rc < NA AND sc >= 0 AND sc < NB *>
              RefineAux(
                NA, NB,
                stepCostFn := stepCostFn,
                minAlign := minAlign, maxAlign := maxAlign,
                ctr := Pair{rc, sc},
                mt := mtBest, m := mBest, totCost := totCost,
                minTotCost := minTotCost,
                c := c,
                cMin := cMin
              )
            END;
            (*
              DebugIntPair("mtBest[0]:", mtBest[0,0], mtBest[0,1]);
              DebugIntPair("mtBest[" & Fmt.Int(mBest-1) & "]: ", mtBest[mBest-1,0],
              mtBest[mBest-1,1]);
              Wr.PutText(stderr,"minimum total cost : " & Fmt.LongReal(totCost) & "\n");
            *)
          END;
          (* Return the best match: *)
          <* ASSERT mBest > 0 *>
          mt := Trim(mtBest, 0, mBest);
        END
      END
    END;
  END Refine;

PROCEDURE ComputeAlignmentRanges(
    NA, NB: NAT;
    READONLY mt: T;
    maxAdjust: NAT;
    VAR (*OUT*) minAlign: INTS;
    VAR (*OUT*) maxAlign: INTS;
  ) =
  (*
    This procedure is used by "Refine". It computes,
    for each position "p", the minimum and maximum alignments
    of valid index pairs for matchings between chains of
    "NA" and "NB" elements, as implied by the initial matching "mt"
    and fattening amount "maxAdjust". *)
  VAR k: NAT := 0;
  BEGIN
    (*
      The "minAlign" and "maxAlign" arrays are basically the
      alignments of the pairs in "mt", shifted by "maxAdjust".
      However each full step of "mt" skips over one position,
      so we must we must interpolate the alignment bounds there.
      We must also extrapolate beyond the ends of "mt".
    *)

    <* ASSERT NUMBER(mt) > 0 *>
    WITH
      minPos = 0 + 0,
      maxPos = (NA-1) + (NB-1),

      m = NUMBER(mt),
      iniPos = mt[0][0] + mt[0][1],
      finPos = mt[m-1][0] + mt[m-1][1],
      iniAlign = mt[0][0] - mt[0][1],
      finAlign = mt[m-1][0] - mt[m-1][1]
    DO
      <* ASSERT minPos = 0 *>
      <* ASSERT LAST(minAlign) = maxPos *>
      <* ASSERT LAST(maxAlign) = maxPos *>
      FOR p := minPos TO maxPos DO
        (* Compute boundary of extrapolated and expanded matching "mt": *)
        IF p <= iniPos THEN
          minAlign[p] := iniAlign - (iniPos - p) - 2*maxAdjust;
          maxAlign[p] := iniAlign + (iniPos - p) + 2*maxAdjust
        ELSIF p >= finPos THEN
          minAlign[p] := finAlign - (p - finPos) - 2*maxAdjust;
          maxAlign[p] := finAlign + (p - finPos) + 2*maxAdjust
        ELSE
          (* Get pair of "mt" at this position: *)
          WHILE mt[k][0] + mt[k][1] < p DO INC(k) END;
          <* ASSERT k < m *>
          WITH
            rk = mt[k][0],
            sk = mt[k][1],
            nxtPos = rk + sk,
            nxtAlign = rk - sk
          DO
            <* ASSERT nxtPos >= p AND nxtPos <= p+1 *>
            minAlign[p] := nxtAlign - 2*maxAdjust;
            maxAlign[p] := nxtAlign + 2*maxAdjust;
            IF nxtPos > p THEN
              (* Full step, skipped over the position "p": *)
              <* ASSERT k > 0 *>
              INC(minAlign[p]);
              DEC(maxAlign[p])
            END
          END
        END;

        (* Clip valid pairs against rectangle "[0..NA-1][0..NB-1]". *)
        WITH
          loAlign = MAX(-p, p - 2*(NB-1)),
          hiAlign = MIN(+p, 2*(NA-1) - p)
        DO
          minAlign[p] := MAX(minAlign[p], loAlign);
          maxAlign[p] := MIN(maxAlign[p], hiAlign);
        END;

        (* Check parity: *)
        <* ASSERT minAlign[p] MOD 2 = p MOD 2 *>
        <* ASSERT maxAlign[p] MOD 2 = p MOD 2 *>
      END
    END
  END ComputeAlignmentRanges;

PROCEDURE RefineAux(
    NA, NB: NAT;
    stepCostFn: StepFunction;
    READONLY minAlign, maxAlign: INTS;
    READONLY ctr: Pair;
    VAR (*IO*) mt: T;
    VAR (*IO*) m: NAT;
    VAR (*IO*) totCost: LONG;
    VAR (*IO*) minTotCost: LONGS;
    VAR (*WORK*) c: CostMatrix;
    VAR (*WORK*) cMin: LONGS;
  ) =
  (*
    Similar to "Refine", but considers only matches that contain the
    pair "ctr" (which must be valid).

    Assumes that the alignments of all valid pairs for a given position "p"
    are the integers in the interval "[ minAlign[p] .. maxAlign[p] ]"
    with same parity as "p"; where "p" ranges from "0+0"
    to "(NA-1)+(NB-1)".

    The parameters "mtBest", "m", "mismatchBest", and "minTotCost"
    must be allocated and initialized by the client.  Before and after
    the call, the best valid matching known is "mt[0..m-1]", and its
    mismatch is "mismatchBest".  The minimum cost of any valid path
    with step count "k" is "minTotCost[k]".

    The procedure updates "mtBest" and "mismatchBest" if (and only if)
    it finds a matching better that the current one.  Similarly, it
    updates "minTotCost[k]" if (and only if) it finds a path of step
    count "k" whose total cost is better than the current value.

    The procedure requires a work vector "cMin" with at least "NA + NB +
    1" elements, and a matrix "c" with at least "NA" rows and "NB"
    columns.
  *)
  BEGIN
    WITH
      (* Range of positions of valid pairs: *)
      minPos = 0 + 0,
      maxPos = (NA-1) + (NB-1),

      (* Position and alignment of mandatory pair: *)
      ctrPos = ctr[0] + ctr[1],
      ctrAlign = ctr[0] - ctr[1]
    DO
      <* ASSERT minPos = 0 *>

      (*
        Initialize the "cMin" array. Element "cMin[p]" is
        the min cost for any matching from the current center to any
        pair "(r,s)" with "r+s = p".

        Note that "cMin[p]" is indexed by the `position' coordinate
        of the distal pair, whereas "minTotCost" is indexed
        by the *step count* of the match.
      *)
      <* ASSERT LAST(cMin) >= (NA-1)+(NB-1) *>
      FOR p := minPos TO maxPos DO cMin[p] := LAST(LONG) END;

      TYPE Step = RECORD cost: LONG; org: Pair END;

      (*
        The procedures "CheapestStep" and "ExtractPath" below
        are used by the dynamic programming algorithm to find
        the optimal path that starts at the "ctr" pair and
        extends in the direction "dir".

        Both procedures assume that the valid pairs for the current
        pass have positions between "ctrPos" and "finPos" (which may
        be greater or less than "ctrPos"), and that "pos" gets
        incremented by "dir" or "2*dir" for each step along the path.
      *)

      PROCEDURE CheapestStep(READONLY dst: Pair; dir: Direction; finPos: NAT): Step =
        (*
          The inner loop of dynamic programming algorithm.  This
          procedure computes the cheapest step ending at pair "dst"
          and coming from one of the three neighboring pairs in the
          direction "-dir".
        *)
        VAR bestCost: LONG;
            bestOrg: Pair;
        BEGIN
          <* ASSERT dir # 0 *>
          WITH
            r = dst[0],
            s = dst[1],
            r1 = r - dir,
            s1 = s - dir
          DO
            (* Sanity checks about the pair "dst = (r,s)": *)
            WITH
              dstPos = r + s,
              dstAlign = r - s
            DO
              <* ASSERT (dstPos - ctrPos)*dir > 0 *>
              <* ASSERT (finPos - dstPos)*dir >= 0 *>
              <* ASSERT minAlign[dstPos] <= dstAlign AND dstAlign <= maxAlign[dstPos] *>
              <* ASSERT ABS(dstAlign - ctrAlign) <= ABS(dstPos - ctrPos) *>
            END;

            bestCost := LAST(LONG);

            (* Try full step from "(r1,s1)": *)
            WITH
              orgPos = r1 + s1,
              orgAlign = r1 - s1
            DO
              IF (orgPos - ctrPos)*dir >= 0
              AND minAlign[orgPos] <= orgAlign AND orgAlign <= maxAlign[orgPos]
              AND ABS(orgAlign - ctrAlign) <= ABS(orgPos - ctrPos) THEN
                WITH
                  org = Pair{r1,s1},
                  cost = c[r1,s1] + stepCostFn(org, dst)
                DO
                  IF cost < bestCost THEN bestCost := cost; bestOrg := org END;
                END
              END
            END;

            (* Try half-step from "(r1,s)": *)
            WITH
              orgPos = r1 + s,
              orgAlign = r1 - s
            DO
              IF minAlign[orgPos] <= orgAlign AND orgAlign <= maxAlign[orgPos]
              AND ABS(orgAlign - ctrAlign) <= ABS(orgPos - ctrPos) THEN
                WITH
                  org = Pair{r1,s},
                  cost = c[r1,s] + stepCostFn(org, dst)
                DO
                  IF cost < bestCost THEN bestCost := cost; bestOrg := org END;
                END
              END
            END;

            (* Try half-step from "(r,s1)": *)
            WITH
              orgPos = r + s1,
              orgAlign = r - s1
            DO
              IF minAlign[orgPos] <= orgAlign AND orgAlign <= maxAlign[orgPos]
              AND ABS(orgAlign - ctrAlign) <= ABS(orgPos - ctrPos) THEN
                WITH
                  org = Pair{r,s1},
                  cost = c[r,s1] + stepCostFn(org, dst)
                DO
                  IF cost < bestCost THEN bestCost := cost; bestOrg := org END;
                END
              END;
            END;

            <* ASSERT bestCost < LAST(LONG) *>
            RETURN Step{cost := bestCost, org := bestOrg}
          END
        END CheapestStep;

      PROCEDURE ExtractMinCostPath(
          dst: Pair;
          dir: Direction;
          finPos: NAT;
        ) =
        (*
          Extracts result of the dynamic programming pass in direction "dir".

          Extracts from "c" the cheapest path from the "ctr" pair to
          the specified "dst" pair.  The path is stored into the "mt"
          array, starting with "mt[m]", and reversed as needed to
          get the pairs in the correct order.
        *)
        VAR mIni: NAT := m;
        BEGIN
          (* 
            Wr.PutText(stderr, "ctr = (" & Fmt.Int(ctr[0]) & "," & Fmt.Int(ctr[1]) & ")\n");
            Wr.PutText(stderr, "dir = " & Fmt.Int(dir) & "\n");
          *)
          <* ASSERT dir # 0 *>
          WHILE dst # ctr DO
            (* Sanity checks about the pair "dst": *)
            WITH
              r = dst[0],
              s = dst[1],
              dstPos = r + s,
              dstAlign = r - s
            DO
              (* 
                Wr.PutText(stderr, "dst = (" & Fmt.Int(dst[0]) & "," & Fmt.Int(dst[1]) & ")\n");
              *)
              <* ASSERT (dstPos - ctrPos)*dir > 0 *>
              <* ASSERT (finPos - dstPos)*dir >= 0 *>
              <* ASSERT minAlign[dstPos] <= dstAlign AND dstAlign <= maxAlign[dstPos] *>
              <* ASSERT ABS(dstAlign - ctrAlign) <= ABS(dstPos - ctrPos) *>
            END;

            (* OK, store pair in "mt" and move to previous one: *)
            mt[m] := dst; INC(m);
            dst := CheapestStep(dst, dir, finPos).org
          END;

          IF dir = +1 THEN
            (* Reverse pairs: *)
            WITH mx = m - mIni, mtx = SUBARRAY(mt, mIni, mx) DO
              FOR i := 0 TO NUMBER(mtx) DIV 2 - 1 DO
                WITH j = LAST(mtx) - i DO
                  VAR t := mtx[i]; BEGIN mtx[i] := mtx[j]; mtx[j] := t END
                END
              END
            END
          END
        END ExtractMinCostPath;

      PROCEDURE FindBestPairInLayer(p: NAT): Pair =
        (*
          Finds the valid pair with position "p" that
          has minimum cost in the "c" array. Breaks ties
          towards the center of the valid alignment range
          (i.e. closer to the original matching).
        *)
        VAR
          bestCost: LONG := LAST(LONG);
          bestPair: Pair;
          bestShift: NAT := LAST(NAT);
        BEGIN
          WITH
            aLo = MAX(minAlign[p], ctrAlign - ABS(p - ctrPos)),
            aHi = MIN(maxAlign[p], ctrAlign + ABS(p - ctrPos)),
            aMd = (minAlign[p] + maxAlign[p]) DIV 2
          DO
            <* ASSERT aLo MOD 2 = p MOD 2 *>
            <* ASSERT aHi MOD 2 = p MOD 2 *>
            FOR a := aLo TO aHi BY 2 DO
              WITH
                r = (p+a) DIV 2,
                s = (p-a) DIV 2,
                cost = c[r,s],
                shift = ABS(a - aMd)
              DO
                IF cost < bestCost
                OR (cost = bestCost AND shift < bestShift) THEN
                  bestPair := Pair{r,s};
                  bestCost := cost; bestShift := shift;
                END
              END
            END;
            <* ASSERT bestCost < LAST(LONG) *>
            RETURN bestPair
          END
        END FindBestPairInLayer;

      BEGIN
        (* Bi-directional dynamic programming: *)
        c[ctr[0],ctr[1]]:= 0.0d0;
        cMin[ctrPos] := 0.0d0;
        FOR dir := -1 TO +1 BY 2 DO
          WITH
            finPos = ARRAY Direction OF NAT{ minPos, 0, maxPos } [dir]
          DO
            FOR p := ctrPos+dir TO finPos BY dir DO
              WITH
                aLo = MAX(minAlign[p], ctrAlign - ABS(p - ctrPos)),
                aHi = MIN(maxAlign[p], ctrAlign + ABS(p - ctrPos))
              DO
                <* ASSERT aLo MOD 2 = p MOD 2 *>
                <* ASSERT aHi MOD 2 = p MOD 2 *>
                FOR a := aLo TO aHi BY 2 DO
                  WITH
                    r = (p+a) DIV 2,
                    s = (p-a) DIV 2,
                    cost = CheapestStep(Pair{r,s}, dir, finPos).cost
                  DO
                    c[r,s] := cost;
                    cMin[p] := MIN(cMin[p], cost)
                  END
                END
              END
            END
          END
        END;

        (* Look for minimum mismatch path, updating "mt", "mismatch", etc.: *)
        VAR
          bestPosLo, bestPosHi: NAT;
          bestTotCost: LONG := totCost;
        BEGIN
          FOR pLo:= minPos TO ctrPos DO
            FOR pHi := ctrPos TO maxPos DO
              WITH
                tc = cMin[pLo] + cMin[pHi],
                stepCount = pHi - pLo
              DO
                minTotCost[stepCount] := MIN(minTotCost[stepCount], tc);
                IF tc < bestTotCost THEN
                  bestTotCost := tc; bestPosLo := pLo; bestPosHi := pHi;
                END
              END
            END;
          END;

          (* DebugIntPair("bestPosLo,bestPosHi:", bestPosLo,bestPosHi); *)
          (* If we found a better matching, let's extract it: *)
          IF bestTotCost < totCost THEN
            WITH
              org = FindBestPairInLayer(bestPosLo),
              dst = FindBestPairInLayer(bestPosHi)
            DO
              (* Extract matching: *)
              m := 0;
              ExtractMinCostPath(org, -1, minPos);
              mt[m] := ctr; INC(m);
              ExtractMinCostPath(dst, +1, maxPos)
            END;
            totCost := bestTotCost
          END;
        END
      END
    END
  END RefineAux;

PROCEDURE AbsPairClip(p: Pair; READONLY s: PZSegment.Pair): Pair =
  BEGIN
    RETURN Pair{AbsIndexClip(p[0], s[0]), AbsIndexClip(p[1], s[1])}
  END AbsPairClip;

PROCEDURE RelPairClip(p: Pair; READONLY s: PZSegment.Pair): Pair =
  BEGIN
    RETURN Pair{RelIndexClip(p[0], s[0]), RelIndexClip(p[1], s[1])}
  END RelPairClip;

PROCEDURE Trim(READONLY mt: T; start, count: NAT): REF T =
  BEGIN
    WITH r = NEW(REF T, count) DO
      r^ := SUBARRAY(mt, start, count);
      RETURN r
    END
  END Trim;

PROCEDURE Adjust(
    READONLY mt: T;
    READONLY sOld: PZSegment.Pair;
    READONLY sNew: PZSegment.Pair;
  ): REF T =

  VAR mNewR: REF T;   (* The new matching *)
      kNew: NAT; (* Number of new pairs in "mNewR^ *)
      stop: BOOL;  (* TRUE after first interruption. *)

  PROCEDURE EnsureSpace(VAR mR: REF T; k: NAT) =
    BEGIN
      IF mR = NIL THEN
        mR := NEW(REF T, k)
      ELSIF k > LAST(mR^) THEN
        WITH n = NUMBER(mR^), mT = NEW(REF T, MAX(k+1, 2*n)) DO
          SUBARRAY(mT^,0,n)  := mR^; mR := mT
        END
      END
    END EnsureSpace;

  BEGIN
    kNew := 0;
    stop := FALSE;
    mNewR := NEW(REF T, NUMBER(mt));
    FOR kOld := 0 TO LAST(mt) DO
      WITH
        rOld = mt[kOld],       (* Pair relative to "sOld" *)
        p = AbsPairClip(rOld, sOld), (* Pair, absolute *)
        rNew = RelPairClip(p, sNew)  (* Pair, relative to "sNew" *)
      DO
        IF rNew[0] = LAST(NAT) OR rNew[1] = LAST(NAT) THEN
          (* Pair is outside "sNew": *)
          IF kNew > 0 THEN stop := TRUE END
        ELSE
          (* Pair lies inside "sNew" *)
          IF stop THEN RETURN NIL END;
          (* Consistency check: *)
          IF kNew > 0 THEN
            WITH aNew = mNewR[kNew-1] DO
              <* ASSERT aNew[0] < rNew[0] *>
              <* ASSERT aNew[1] < rNew[1] *>
              <* ASSERT aNew[0] + 1 >= rNew[0] *>
              <* ASSERT aNew[1] + 1 >= rNew[1] *>
            END
          END;
          EnsureSpace(mNewR, kNew);
          mNewR^[kNew] := rNew;
          INC(kNew)
        END;
      END
    END;
    IF kNew = 0 THEN RETURN NIL END;
    WITH mR = NEW(REF T, kNew) DO
      mR^ := SUBARRAY(mNewR^, 0, kNew);
      RETURN mR
    END
  END Adjust;

PROCEDURE MostPerfect(n0, n1: NAT): REF T =
  BEGIN
    <* ASSERT n0 > 0 *>
    <* ASSERT n1 > 0 *>
    WITH
      nPairs = MAX(n0, n1),
      rm = NEW(REF T, nPairs), mt = rm^
    DO
      DoMostPerfect(n0, n1, mt);
      RETURN rm
    END
  END MostPerfect;

PROCEDURE DoMostPerfect(n0, n1: NAT; VAR mt: T) =
  BEGIN
    <* ASSERT n0 > 0 *>
    <* ASSERT n1 > 0 *>
    WITH
      nPairs = NUMBER(mt)
    DO
      <* ASSERT nPairs = MAX(n0,n1) *>
      IF nPairs = 1 THEN
        mt[0] := Pair{0,0}
      ELSE
        WITH
          fn0 = FLOAT(n0-1, LONG),
          fn1 = FLOAT(n1-1, LONG),
          fnp = FLOAT(nPairs-1, LONG)
        DO
          FOR i := 0 TO nPairs-1 DO
            WITH
              fi = FLOAT(i,LONG),
              p0 = ROUND((fn0 * fi)/fnp),
              p1 = ROUND((fn1 * fi)/fnp)
            DO
              mt[i] := Pair{p0, p1}
            END
          END
        END
      END
    END
  END DoMostPerfect;

PROCEDURE RemoveDanglingEnds(READONLY mt: T): REF T =
  VAR ini, fin: NAT;
  BEGIN
    ini := 0; fin := LAST(mt);
    WITH
      ix = mt[0][0],
      iy = mt[0][1]
    DO
      WHILE (ini < fin) AND (
          (mt[ini][0] = ix AND mt[ini+1][0] = ix) OR
          (mt[ini][1] = iy AND mt[ini+1][1] = iy)
        )
      DO
        INC(ini)
      END
    END;
    WITH
      ix = mt[LAST(mt)][0],
      iy = mt[LAST(mt)][1]
    DO
      WHILE (ini < fin) AND (
          (mt[fin][0] = ix AND mt[fin-1][0] = ix) OR
          (mt[fin][1] = iy AND mt[fin-1][1] = iy)
        )
      DO
        DEC(fin)
      END
    END;
    RETURN Trim(mt, ini, fin-ini+1)
  END RemoveDanglingEnds;

PROCEDURE RemoveUnmatchedEnds(
    READONLY mt: T;
    isMatched: StepPredicate;
  ): REF T =

  VAR ini, fin: NAT;
  BEGIN
    ini := 0; fin := LAST(mt);
    WHILE ini < fin AND NOT isMatched(mt[ini], mt[ini+1]) DO
      INC(ini)
    END;
    WHILE ini < fin AND NOT isMatched(mt[fin-1], mt[fin]) DO
      DEC(fin)
    END;
    RETURN Trim(mt, ini, fin-ini+1)
  END RemoveUnmatchedEnds;

PROCEDURE Length(
    READONLY mt: T;
    lengthFn: StepMeasure;
  ): LONG =
  VAR tot : LONG := 0.0d0;
  BEGIN
    FOR i := 0 TO LAST(mt)-1 DO
      WITH p0 = mt[i], p1 = mt[i+1] DO
        tot := tot + lengthFn(p0, p1)
      END
    END;
    RETURN tot
  END Length;

PROCEDURE Sum(
    READONLY mt: T;
    stepFn: StepFunction;
  ): LONG =
  VAR tot : LONG := 0.0d0;
  BEGIN
    FOR i := 0 TO LAST(mt)-1 DO
      WITH p0 = mt[i], p1 = mt[i+1] DO
        tot := tot + stepFn(p0, p1)
      END
    END;
    RETURN tot
  END Sum;

PROCEDURE Integral(
    READONLY mt: T;
    lengthFn: StepMeasure;
    stepFn: StepFunction;
  ): LONG =
  VAR tot : LONG := 0.0d0;
  BEGIN
    FOR i := 0 TO LAST(mt)-1 DO
      WITH p0 = mt[i], p1 = mt[i+1] DO
        tot := tot + stepFn(p0, p1)*lengthFn(p0, p1)
      END
    END;
    RETURN tot
  END Integral;

PROCEDURE MatchedStepCount(
    READONLY mt: T;
    isMatched: StepPredicate;
  ): NAT =
  VAR tot: NAT := 0;
  BEGIN
    FOR i := 0 TO LAST(mt)-1 DO
      WITH p0 = mt[i], p1 = mt[i+1] DO
        IF isMatched(p0, p1) THEN INC(tot) END
      END
    END;
    RETURN tot
  END MatchedStepCount;

PROCEDURE MatchedLength(
    READONLY mt: T;
    lengthFn: StepMeasure;
    isMatched: StepPredicate;
  ): LONG =
  VAR tot: LONG := 0.0d0;
  BEGIN
    FOR i := 0 TO LAST(mt)-1 DO
      WITH p0 = mt[i], p1 = mt[i+1] DO
        IF isMatched(p0, p1) THEN
          tot := tot + lengthFn(p0, p1)
        END
      END
    END;
    RETURN tot
  END MatchedLength;

CONST DebugOverlap = FALSE;

PROCEDURE Overlap(
    READONLY a, b: T;
    READONLY aSeg, bSeg: PZSegment.Pair;
    maxAdjust: NAT;
  ): NAT =
  
  VAR N: NAT := 0;
      tot: Pair;  (* Total samples in each contour. *)
      bIni: Pair; (* Start of "a" in the flipped toroidal grid. *)
      aIni: Pair; (* Start of "b" in the flipped toroidal grid. *)
      (*
        The two matchings are viewed as paths on the toroidal
        grid, relative to "aIni" and "bIni", respectively.
        The points "aIni" and "bIni" have been mirrored, if
        necessary, so that the paths are increasing along both axes.
      *)
  
  PROCEDURE Tally(ai, af, bi, bf: NAT) =
    (*
      Adds to "N" twice the length of steps "aIni+a[ak]->aIni+a[ak+1]" that 
      are within "maxAdjust" some step "bIni+b[bk]->bIni+b[bk+1]", 
      in the X-Y direction, where "ak IN [ai..af-1]" and "bk IN [bi..bf-1]". 
    *)
    BEGIN
      <* ASSERT ai < af *>
      <* ASSERT bi < bf *>
      IF DebugOverlap THEN
        Wr.PutText(stderr, 
          "[ " & FI(ai,1) & "-" & FI(af,1) & "" & FI(bi,1) & "-" & FI(bf,1)
        )
      END;
      IF NOT BoxesOverlap(a[ai], a[af], b[bi], b[bf]) THEN 
        (* Do nothing *)
      ELSIF af - ai >= bf - bi AND af - ai >= 2 THEN 
        (* The "a" matching is longer and has at least two steps; split it: *)
        WITH
          am = (ai + af) DIV 2
        DO
          <* ASSERT ai < am AND am < af *>
          Tally(ai, am, bi, bf);
          Tally(am, af, bi, bf);
        END
      ELSIF bf - bi >= af - ai AND bf - bi >= 2 THEN 
        (* The "b" matching is longer and has at least two steps; split it: *)
        WITH
          bm = (bi + bf) DIV 2
        DO
          <* ASSERT bi < bm AND bm < bf *>
          Tally(ai, af, bi, bm);
          Tally(ai, af, bm, bf);
        END
      ELSE
        (*
          Both "a" and "b" are reduced to a single step.  Let "B" be
          the step "b", smeared diagonally by "maxAdjust".  Since "B"
          is convex, it suffices to check whether both endpoints of
          "a" are in "B".
          
          We know that B overlaps non-trivially the bounding box of
          "a".  Moreover the only integer points in "B" are also in
          the bloating of the two endpoints.
        *)
        WITH ov = StepOverlap(a[ai], a[af], b[bi], b[bf]) DO
          IF ov > 0 THEN 
            IF DebugOverlap THEN
              Wr.PutText(stderr, " " & "+" & FI(ov,1) & " ")
            END;
            INC(N, ov)
          END
        END
      END;
      IF DebugOverlap THEN Wr.PutText(stderr, "]") END;
    END Tally;
    
  PROCEDURE BoxesOverlap(READONLY aiP, afP, biP, bfP: Pair): BOOL =
    (*
      Returns true if the bounding box of the path
      "aIni+aiP->aIni+afP" intersects the bounding box of
      "bIni+biP->bIni+bfP", smeared diagonally by "maxAdjust". The
      paths are assumed to be on the toroidal grid, but it is required
      that "afP >= aiP" and "bfP >= biP".
      
      The boxes are assumed not to include the corners corresponding
      to the end-pairs of the matching. *)
    VAR n, d: Pair; 
    BEGIN
      (*
        Let "na[j]" and "nb[j] be the extents of the two paths along
        axis "j", and "Q = maxAdjust".  We need to know whether there
        are real vectors "va" in "[0 _ na[0]][0 _ na[1]]" and "vb" in
        "[0 _ nb[0]][0 _ nb[1]]", minus the two extreme corners, and a
        vector "vc" in "{ (-k,k) : k in [-Q _ +Q] }", such that
        "(aIni+aiP+va) - (bIni+bfP-vb) = vc" modulo the two chain
        lengths "tot[0],tot[1]".
                
        This condition is equivalent to asking whether there are real
        vectors "va,vb,vc" in the ranges above and integers "r,s" such
        that
        
          "va+vb+vc+(Q,Q) = (bIni+bfP)-(aIni+aiP)+(Q,Q) + (r*tot[0],s*tot[1])"
          
        over the Cartesian integer grid.  The set of all vectors
        "(x,y)" on the left-hand side is the subset of the box
        "C = [0 _ na[0]+nb[0]+2*Q][0 _ na[1]+nb[1]+2*Q]" that
        satisfies the `diagonal' inequalities
        
          "x+y > 2*Q"
          "x+y < na[0]+na[1]+nb[0]+nb[1]+2*Q"
        
        So we need to check only whether this region contains
        any vectors congruent to "d = (bIni+bfP)-(aIni+aiP)+(Q,Q)"
      *)
      (* Compute the box parameters: *)
      FOR j := 0 TO 1 DO
        WITH m = tot[j] DO
          <* ASSERT aiP[j] <= afP[j] AND afP[j] < aiP[j]+m *>
          <* ASSERT biP[j] <= bfP[j] AND bfP[j] < biP[j]+m *>
          n[j] := (afP[j] - aiP[j]) + (bfP[j] - biP[j]) + 2*maxAdjust;
          d[j] := ((bIni[j] + bfP[j]) - (aIni[j] + aiP[j]) + maxAdjust) MOD m;
          (* Coarse check for separation: *)
          IF d[j] > n[j] THEN 
            (* The point "d" lies outside the box "C"; it follows that *)
            (* all points congruent to "d" also are outside "C". *)
            RETURN FALSE
          END
        END
      END;
      (*
        Now we know that the point "d" lies inside the 
        box "C". We still have to check whether for some
        "r" and "s" the point "d+(r*tot[0],s*tot[1])" 
        lies in "C" and satisfies the diagonal inequalities.
        Since we expect "Q" to be small, and "n[j]" to be 
        hardly greater than "2*tot[j]", we can do it by 
        exhaustive enumeration over "r" and "s":
      *)
      WITH
        loD = maxAdjust+maxAdjust,
        hiD = n[0] + n[1] - loD
      DO
        FOR r := 0 TO (n[0] - d[0]) DIV tot[0] DO
          FOR s := 0 TO (n[1] - d[1]) DIV tot[1] DO
            WITH
              x = d[0] + r*tot[0],
              y = d[1] + r*tot[1]
            DO
              <* ASSERT x <= n[0] *>
              <* ASSERT y <= n[1] *>
              IF x+y > loD AND x+y < hiD THEN RETURN TRUE END
            END
          END
        END
      END;
      RETURN FALSE
    END BoxesOverlap;
    
  PROCEDURE StepOverlap(READONLY aiP, afP, biP, bfP: Pair): NAT =
    (*
      Returns the fraction (in multiples of a half-step) 
      of the step "a = aIni+aiP->aIni+afP" that lies
      at most "maxAdjust" away diagonally from the step
      "b = bIni+biP->bIni+bfP".  It is given that the bounding
      box of "a" intersects the diagonally smeared box of "b". *)
    VAR na, nb, n, d: Pair;
        NS: NAT;
    BEGIN
      (* 
        To simplify the logic, we reduce the problem to the steps
        "A = (Q,Q)->(Q,Q)+na" and "B = d->d+nb", where "d" is in the 
        fundamental rectangle of the toroidal grid.  Note that
        the smeared box of "A" consist of the points "(x,y)" of 
        the real rectangle "C = [0 _ na[0]+2*Q][0 _ na[1]+2*Q]"
        on the Cartesian plane that satisfies the `diagonal' inequalities
        
          "x+y > 2*Q"
          "x+y < na[0]+na[1]+2*Q"

      *)
      FOR j := 0 TO 1 DO
        WITH m = tot[j] DO
          na[j] := afP[j] - aiP[j];
          <* ASSERT 0 <= na[j] AND na[j] <= 1 *>
          n[j] := na[j] + maxAdjust+maxAdjust;
          nb[j] := bfP[j] - biP[j];
          <* ASSERT 0 <= nb[j] AND nb[j] <= 1 *>
          d[j] := ((bIni[j] + biP[j]) - (aIni[j] + aiP[j]) + maxAdjust) MOD m;
          (* Coarse check for separation: *)
          IF d[j] + nb[j] > n[j] THEN
            (* The "b" step ends outside the "C" rectangle. *)
            (* It follows that any of its translates intersects "C" *)
            (* at most in two points. *)
            RETURN 0
          END
        END
      END;
      (* 
        Now we know that "B" lies inside the closed rectangle "C".
        We must now tally the intersection of the smeared box
        of "A" with all congruent copies of step "B", namely
        all steps "u+d->u+d+nb" where "u=(r*tot[0],s*tot[1])".
        We only need to consider those translates that intersect
        non-trivially the closed rectangle "C".
        Since "d[j]" lies in "[0..tot[j]-1]", we need not consider
        "r<0", or "r" such that "r*tot[0]+d[0]+nb[0] > na[0]+2*Q";
        and ditto for "s".
      *)
      WITH
        aID = maxAdjust+maxAdjust,
        aFD = aID + na[0] + nb[1]
      DO
        NS := 0;
        FOR r := 0 TO (na[0]+maxAdjust+maxAdjust-d[0]-nb[0]) DIV tot[0] DO
          FOR s := 0 TO (na[1]+maxAdjust+maxAdjust-d[1]-nb[1]) DIV tot[1] DO
            WITH
              bI0 = d[0] + r*tot[0], 
              bF0 = bI0 + nb[0],
              bI1 = d[1] + r*tot[1], 
              bF1 = bI1 + nb[1],
              bID = bI0 + bI1, 
              bFD = bF0 + bF1
            DO
              <* ASSERT bF0 <= n[0] *>
              <* ASSERT bF1 <= n[1] *>
              NS := NS + MAX(0, MIN(aFD,bFD) - MAX(aID,bID))
            END
          END
        END;
        RETURN NS
      END
    END StepOverlap;      
    
  BEGIN
    IF aSeg[0].cvx # bSeg[0].cvx OR aSeg[1].cvx # bSeg[1].cvx THEN RETURN 0 END;
    IF aSeg[0].rev # bSeg[0].rev OR aSeg[1].rev # bSeg[1].rev THEN RETURN 0 END;
    <* ASSERT aSeg[0].tot = bSeg[0].tot AND aSeg[1].tot = bSeg[1].tot *>

    (* Flip the problem to account for the "rev" flags: *)
    FOR j := 0 TO 1 DO
      tot[j] := aSeg[j].tot;
      WITH m = tot[j] DO
        IF NOT aSeg[j].rev THEN
          aIni[j] := aSeg[j].ini MOD m;
          bIni[j] := bSeg[j].ini MOD m;
        ELSE
          aIni[j] := (m - (aSeg[j].ini + aSeg[j].ns - 1)) MOD m;
          bIni[j] := (m - (bSeg[j].ini + bSeg[j].ns - 1)) MOD m;
        END;
      END;
    END;
    N := 0;
    WITH 
      aFin = Pair{aIni[0] + aSeg[0].ns-1, aIni[1] + aSeg[1].ns-1},
      bFin = Pair{bIni[0] + bSeg[0].ns-1, bIni[1] + bSeg[1].ns-1}
    DO
      IF DebugOverlap THEN
        Wr.PutText(stderr, " " & 
          "(" & FI(aIni[0],1) & "," & FI(aIni[1],1) & ")->" &
          "(" & FI(aFin[0],1) & "," & FI(aFin[1],1) & ")  " &
          "(" & FI(bIni[0],1) & "," & FI(bIni[1],1) & ")->" &
          "(" & FI(bFin[0],1) & "," & FI(bFin[1],1) & ")"
        )
      END;
    END;
    IF LAST(a) > 0 AND LAST(b) > 0 THEN 
      Tally(0, LAST(a), 0, LAST(b))
    END;
    IF DebugOverlap THEN Wr.PutText(stderr, "=" & Fmt.Int(N) & "\n") END;
   RETURN N
  END Overlap;

PROCEDURE Pack(READONLY mt: T): REF PackedT =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    WITH
      np = NUMBER(mt),
      rp = NEW(REF PackedT), pm = rp^,
      wr = NEW(TextWr.T).init()
    DO
      Wr.PutChar(wr, '|');
      FOR i := 1 TO np-1 DO
        WITH
          ri = mt[i][0], ra = mt[i-1][0],
          si = mt[i][1], sa = mt[i-1][1]
        DO
          IF ri = ra+1 AND si = sa THEN
            Wr.PutChar(wr, '/')
          ELSIF ri = ra AND si = sa+1 THEN
            Wr.PutChar(wr, '\\')
          ELSIF ri = ra+1 AND si = sa+1 THEN
            Wr.PutChar(wr, '|')
          ELSE
            Wr.PutText(stderr, "bad matching: ");
            Wr.PutText(stderr, Fmt.Int(ra) & " " & Fmt.Int(ri) & " ");
            Wr.PutText(stderr, Fmt.Int(sa) & " " & Fmt.Int(si) & "\n");
            <* ASSERT FALSE *>
          END
        END
      END;
      pm := PackedT{
        np := np,
        ini := mt[0],
        fin := mt[np-1],
        steps := TextWr.ToText(wr)
      };
      RETURN rp
    END
  END Pack;

PROCEDURE PackPerfect(nSamples: NAT): REF PackedT =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    WITH
      rpm = NEW(REF PackedT), pm = rpm^,
      wr = NEW(TextWr.T).init()
    DO
      pm.np := nSamples;
      pm.ini := Pair{0,0};
      pm.fin := Pair{nSamples-1,nSamples-1};
      FOR k := 0 TO nSamples-1 DO Wr.PutChar(wr, '|') END;
      pm.steps := TextWr.ToText(wr);
      RETURN rpm
    END
  END PackPerfect;

PROCEDURE Unpack(READONLY pm: PackedT): REF T =
  BEGIN
    WITH
      np = pm.np,
      rm = NEW(REF T, np), mt = rm^
    DO
      DoUnpack(pm, mt);
      RETURN rm
    END
  END Unpack;

PROCEDURE DoUnpack(READONLY pm: PackedT; VAR mt: T) =
  (*
    Like "Unpack", but stores the result in the array "m",
    which is assumed to have the correct size. *)
  BEGIN
    WITH np = pm.np DO
      <* ASSERT NUMBER(mt) = np *>
      <* ASSERT Text.Length(pm.steps) = np *>
      <* ASSERT Text.GetChar(pm.steps, 0) = '|' *>
      mt[0] := pm.ini;
      FOR i := 1 TO np-1 DO
        WITH c = Text.GetChar(pm.steps, i) DO
          IF c = '/' THEN
            mt[i][0] := mt[i-1][0] + 1; mt[i][1] := mt[i-1][1];
          ELSIF c = '\\' THEN
            mt[i][0] := mt[i-1][0]; mt[i][1] := mt[i-1][1] + 1;
          ELSIF c = '|' THEN
            mt[i][0] := mt[i-1][0] + 1; mt[i][1] := mt[i-1][1] + 1;
          ELSE
            <* ASSERT FALSE *>
          END
        END
      END;
      <* ASSERT mt[np-1] = pm.fin *>
    END
  END DoUnpack;

PROCEDURE WritePacked(wr: Wr.T; READONLY pm: PackedT) =
  <* FATAL Thread.Alerted, Wr.Failure *>
  BEGIN
    WITH np = pm.np DO
      <* ASSERT Text.Length(pm.steps) = np *>
      FPut.Int(wr, np); FPut.Space(wr);
      FPut.Char(wr, '(');
      FOR j := 0 TO 1 DO
        FPut.Int(wr, pm.ini[j]); FPut.Space(wr);
        FPut.Int(wr, pm.fin[j]); FPut.Space(wr); FPut.Space(wr);
      END;
      Wr.PutText(wr, pm.steps);
      FPut.Char(wr, ')');
    END
  END WritePacked;

PROCEDURE ReadPacked(rd: Rd.T): REF PackedT =
  VAR np: NAT;
  CONST ValidChars = SET OF CHAR{ '/', '|', '\\'};
  BEGIN
    np := FGet.Int(rd);
    WITH
      cm = NEW(REF ARRAY OF CHAR, np)^,
      rp = NEW(REF PackedT), pm = rp^
    DO
      pm.np := np;
      FGet.Skip(rd, FGet.Spaces);
      FGet.Match(rd, "(");
      FOR j := 0 TO 1 DO
        pm.ini[j] := FGet.Int(rd);
        pm.fin[j] := FGet.Int(rd);
      END;
      FGet.Skip(rd, FGet.Spaces);
      FGet.Match(rd, "|"); cm[0] := '|';
      FOR i := 1 TO np-1 DO
        cm[i] := FGet.Char(rd);
        <* ASSERT cm[i] IN ValidChars *>
      END;
      FGet.Skip(rd, FGet.Spaces);
      FGet.Match(rd, ")");
      pm.steps := Text.FromChars(cm);
      RETURN rp
    END
  END ReadPacked;

PROCEDURE AdjustPacked(rpm: REF PackedT;
    READONLY oldSeg: PZSegment.Pair;
    READONLY newSeg: PZSegment.Pair;
  ): REF PackedT =

  PROCEDURE DoAdjust(pm: PackedT): REF PackedT =
    (*
      Adjusts "pm" the hard way - unpack,trim,repack. *)
    BEGIN
      WITH
        r = Adjust(Unpack(pm)^, oldSeg, newSeg)
      DO
        IF r = NIL THEN
          RETURN NIL
        ELSE
          RETURN Pack(r^)
        END
      END
    END DoAdjust;

  VAR adjust: ARRAY [0..1] OF INT;
  BEGIN
    IF rpm = NIL THEN RETURN NIL END;
    WITH
      pm = rpm^
    DO
      (*
        Try to determine if the old pairings are contained
        in the new segments.  If that is the case, we don't
        need to unpack/repack the matching. *)
      FOR j := 0 TO 1 DO
        WITH
          os = oldSeg[j],
          ns = newSeg[j],
          is = PZSegment.Meet(oldSeg[j], newSeg[j]),
          tot = os.tot,
          rev = os.rev,
          iniAdv = (is.ini - os.ini) MOD tot,
          iniRet = (is.ini - ns.ini) MOD tot,
          finAdv = (ns.ini + ns.ns - is.ini - is.ns) MOD tot,
          finRet = (os.ini + os.ns - is.ini - is.ns) MOD tot
        DO
          <* ASSERT os.rev = ns.rev *>
          <* ASSERT os.cvx = ns.cvx *>
          <* ASSERT os.tot = ns.tot *>
          IF is.ns = 0 THEN RETURN NIL END;
          IF rev THEN
            IF finAdv = 0 THEN
              IF finRet > pm.ini[j] THEN RETURN DoAdjust(pm) END;
              adjust[j] := - finRet
            ELSE
              adjust[j] := + finAdv
            END;
            IF iniAdv > os.ns - 1 - pm.fin[j] THEN RETURN DoAdjust(pm) END;
          ELSE
            IF iniRet = 0 THEN
              IF iniAdv > pm.ini[j] THEN RETURN DoAdjust(pm) END;
              adjust[j] := - iniAdv
            ELSE
              adjust[j] := + iniRet
            END;
            IF finRet > os.ns - 1 - pm.fin[j] THEN RETURN DoAdjust(pm) END;
          END
        END
      END;
      WITH rpmNew = NEW(REF PackedT), pmNew = rpmNew^ DO
        pmNew.np := pm.np;
        FOR j := 0 TO 1 DO
          pmNew.ini[j] := pm.ini[j] + adjust[j];
          pmNew.fin[j] := pm.fin[j] + adjust[j];
        END;
        pmNew.steps := pm.steps;
        RETURN rpmNew
      END
    END
  END AdjustPacked;

<*UNUSED*>
PROCEDURE DebugIntPair(title: TEXT; r, s: INT) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, title);
    Wr.PutText(stderr, " ");
    Wr.PutText(stderr, Fmt.Int(r));
    Wr.PutText(stderr, " ");
    Wr.PutText(stderr, Fmt.Int(s));
    Wr.PutText(stderr, "\n");
  END DebugIntPair;

PROCEDURE FI(x: NAT; w: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.Int(x), w)
  END FI;

BEGIN
END PZMatch.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
