INTERFACE PZPlot;

(* Contour plotting routines *)
(* Last edited on 1999-11-08 04:02:34 by hcgl *)

IMPORT PSPlot;
IMPORT PZLR3Chain, PZMatch, PZCandidate, PZSegment;
FROM PZTypes IMPORT LONG, NAT, INT, BOOL;

TYPE 
  SegmentPair = PZSegment.Pair;

  Point = ARRAY [0..1] OF LONG;   (* X and Y coordinates *)

(* CURVES, SEGMENTS, MATCHINGS *)

(* The following procedures draw into an already open "PSPlot" file,  *)
(* with the current graphics settings. *)

PROCEDURE Curve(
    f: PSPlot.File; 
    READONLY c: PZLR3Chain.T;
    closed: BOOL := TRUE;
    step: NAT := 1;
    phase: INT := 0;
  );
  (*
    Draws the given chain into file "f", as a polygonal.
    If "step > 1", uses only the first sample, last
    sample, and one sample out of every "step" 
    samples (including "c[phase]"). *)
    
PROCEDURE Dots(
    f: PSPlot.File; 
    READONLY c: PZLR3Chain.T;
    size: REAL := 2.0; 
    step: NAT := 1;
    phase: INT := 0;
  );
  (*
    Draws dots at the points OF the given chain into file "f",
    plotting only one sample out of every "step" samples,
    including "c[phase]" (modulo the array's length).
    The dots' diameter is "size" times the current line width. *)
    
PROCEDURE Match(
    f: PSPlot.File;
    READONLY m: PZMatch.T;
    READONLY vA, vB: PZLR3Chain.T;
    READONLY segA, segB: PZSegment.T;
    step: NAT := 1;
    phaseA, phaseB: INT := 0;
  );
  (*
    Draws into "f" lines connecting the points of "vA" and "vB" 
    that are connected by the matching "m", with the current line
    thickness, color, and style.  The matching is interpreted
    relative to segments "segA" and "segB".
    
    If "step > 1", plots only one every "step" pairs,
    including the pair "(phaseA,phaseB)" *)

PROCEDURE Segment(
    READONLY f: PSPlot.File;
    s: PZSegment.T;    
    READONLY c: PZLR3Chain.T;  (* Mapped chain *)
    closed: BOOL;              (* TRUE considers the curve closed. *)
    step: NAT := 1;            (* Plot every "step" samples. *)
    phase: INT := 0;           (* Phase of dot sampling. *)
  );
  (*
    Plots a segment of a curve, with the current line thickness,
    color, and style.
    
    The segment extends from "c[s.ini]" to "c[s.fin]".  The indices "s.ini"
    and "s.fin" are reduced modulo "NUMBER(c)"; if "s.ini > s.fin", the
    plotted segment is actually "c[0..s.fin]" and "c[s.ini..LAST(c)]".  In
    this case, if "closed" is TRUE, point "c[LAST(c)]" is connected to
    "c[0]".
    
    If "step > 1" plots only one every "step" samples, synchronized
    with "c[phase]". *)

PROCEDURE SegmentDots(
    READONLY f: PSPlot.File;
    s: PZSegment.T;
    READONLY c: PZLR3Chain.T;  (* Mapped chain *)
    size: REAL := 2.0;
    step: NAT := 1;            (* Plot every "step" samples. *)
    phase: INT := 0;           (* Sampling phase *)
  );
  (*
    Draws dots at every "step" samples of the segment "s", with the
    current line color.  The dots' diameter is "size" times the
    current line width.
    
    If "step > 1" plots only one every "step" samples, sybchronized
    with "c[phase]". *)
    

PROCEDURE SegmentLabel(
    READONLY f: PSPlot.File;
    READONLY s: PZSegment.T;
    READONLY c: PZLR3Chain.T;
    label: TEXT;
    size: REAL := 12.0;
    font: TEXT := "Courier";
  );
  (*
    Labels the segment "s" of curve "c". The label is placed near
    the segment, displaced roughly towards the center of the curve. *)
   
PROCEDURE SegmentPointer(
    READONLY f: PSPlot.File;
    READONLY s: PZSegment.T;
    READONLY c: PZLR3Chain.T;
    frac: LONG;
    width: REAL := 4.0; 
    length: REAL := 8.0; 
    inside: BOOL := TRUE;
  );
  (*
    Places an arrowhead pointing to a point "frac" the way from the
    start to the end of the segment "s" of "c", taking into account
    wrap-around and reversal.  Thus "frac=0" means at start, "frac=1"
    means at end.  
    
    The dimensions are relative to the current line width,
    The arrowhead is painted solid with the current line color.
    
    If "inside=TRUE" the arrow is placed inside the
    curve, otherwise it is placed outside pointing in. *)
   
PROCEDURE SegmentBrackets(
    READONLY f: PSPlot.File;
    READONLY s: PZSegment.T;
    READONLY c: PZLR3Chain.T;
    width: REAL := 8.0; 
    length: REAL := 4.0; 
  );
  (*
    Draws start and finish markers at both ends of segment "s" of "c",
    taking into account wrap-around and reversal.
    
    The dimensions are relative to the current line width.
    The brackets are drawn with the current line style and color. *)
   
(* SCALE ICONS *)

(*
  The following procedures draw pictorial representations of the 
  filtering scale: the graph of the kernel, a circle with critical
  or characteristic radius, a scale bar, etc.. 
  
  The KernelGraph and CriticalCircle procedure draw the icon within a
  rectangle of specified "width" and "height", with its bottom side
  centered on the point "(x,y)".  All these dimensions are in units of
  the current coordinate system.  The icon is drawn with the specified
  "lineWidth" (in mm), and the current line style and color.  The line
  drawing is supressed in "lineWidth" is zero.

  If "dotStep" and "dotSize" are positive, the procedure also plots
  round dots along the graph, spaced "dotStep" in the horizontal
  direction, using the current fill color.  The dots' diameter will be 
  "dotSize" times the current line width. *)

PROCEDURE GaussianKernel(
    f: PSPlot.File;
    lambda: LONG;
    x, y: LONG;
    width: LONG;
    height: LONG;
    lineWidth: REAL;
    dotStep: LONG := 0.0d0;
    dotSize: REAL := 0.0;
  );
  (*
    Draws the graph of a gaussian with standard deviation "lambda",
    clipped to the specified "width" and scaled to fit 
    the specified "height".  The origin of the graph's coordinate system
    will be at "(x,y)". *)
   
PROCEDURE CriticalCircle(
    f: PSPlot.File;
    radius: LONG;
    x, y: LONG;
    width: LONG;
    height: LONG;
    lineWidth: REAL;
    dotStep: LONG := 0.0d0;
    dotSize: REAL := 0.0;
  );
  (*
    Draws a circle with specified "radius".  The circle is centered in
    the specified rectangular window if possible; otherwise only the
    top part of it is shown, clipped to the rectangle. *)
   
PROCEDURE ScaleBar(
    f: PSPlot.File;
    lambda: LONG;
    count: NAT;
    x, y: LONG;
    lineWidth: REAL;
    tickLength: REAL;
  );
  (*
    Draws a horizontal H-bar of length "count*lambda", with ticks 
    every "lambda", centered at the point "(x,y)".  Uses the current
    line style and color. *)

(* ALL-IN-ONE PROCEDURES *)

(* The following procedures use fixed colors, line widths, and styles: *)

PROCEDURE Candidate(
    f: PSPlot.File;
    READONLY cand: PZCandidate.T;
    READONLY c0, c1: PZLR3Chain.T;  (* Chains, already mapped; *)
    whole: BOOL;
    closed: BOOL;
    dots: BOOL;
    pairing: BOOL;
    labels: BOOL;
    pointers: BOOL;
    step: NAT := 1;
  );
  (*
    Plots a candidate, consisting of two segments of the given curves
    "c0" and "c1" (which must have been translated and rotated by
    the client as desired).
    
    The matched segments are drawn in thick red and blue.  The segments may
    wrap around the end of the array, as explained in the "Segment"
    procedure above.
    
    If "whole" is TRUE, the unmatched parts of the contours are drawn
    in thin black.
    
    If "closed" is TRUE the last point of each curve is connected to
    the first.
    
    If "dots" is TRUE then dots are drawn at the sample positions.
    
    If "pairing" is true and the matching "cand.pm" is not NIL, the 
    pairs of the latter are drawn in thin gray. 
    
    IF "labels" is TRUE, the number of each contour (as specified
    by "cand") is drawn somewhere near its center.
    
    If "pointers" is true, draws arrowheads pointing at the 
    start and end of each segment.
    
    If "step > 1", plots only one every "step" points of each
    contour. *)
  
PROCEDURE HalfCandidate(
    f: PSPlot.File;
    s: PZSegment.T;
    READONLY c: PZLR3Chain.T;  (* Chain, already mapped; *)
    color: PSPlot.Color;
    whole: BOOL;
    closed: BOOL;
    dots: BOOL;
    label: BOOL;
    pointers: BOOL;
    step: NAT := 1;
  );
  (* 
    Draws one half of a candidate (minus the pairing),
    as above: the segment "s" in the specified color,
    the rest of the curve "c" in black. *)

TYPE
  ScaleIconStyle = { Kernel, Circle };

PROCEDURE ScaleIcon(
    f: PSPlot.File; 
    filterLambda: LONG;
    filterDotStep: LONG;
    style: ScaleIconStyle
  );
  (*
    Plots an iconic indication of filtering scale, at the 
    lower left corner of the current drawing of "f".
    
    The "Kernel" style icon is the graph of a Gaussian with half-width
    "3*filterLambda", above a scale bar of width "2*lambda" with ticks
    every "lambda".
    
    The "Circle" style icon is a circle with radius 
    "sqrt(e/2)*filterLambda", above a scale bar of width "lambda".
    
    Each icon element will be clipped or supressed if it 
    would be too big compared to the drawing size.
    
    If "filterDotStep" is positive, the procedure also 
    puts down dots with that spacing along the graph or circle.
  *)

(* DYNAMIC PROGRAMMING GRAPH *)

(*
  The following procedures generate illustrations of the "dynamic
  programming" algorithm.  
  
  All procedures draw or assume drawn the dynamic programming graph,
  in the form of a rectangular grid defined by segments "s[0]" and
  "s[1]", with cells of side "cellSize".
  
  The plot will be positioned assuming that the reference grid defined
  by segments "sRef[0]" and "sRef[1]" has its lower left corner at the
  point "(h,h)" where "h = cellSize/2".  The plot will be
  clipped by a rectangle that contains the reference grid surrounded
  by a safety margin of "h".
  
  The procedures that take a match "m" interpret it relative to the
  segment "s", and plot the result relative to "sRef".
  
  The procedures do not take into account that the chains are
  periodic; the segments are interpreted literaly, without reducing
  them modulo "s.tot". Use "TryAllShifts" to get wrap-around plots. *)

PROCEDURE SegSegGrid(
    READONLY f: PSPlot.File;
    READONLY s: SegmentPair; 
    READONLY sRef: SegmentPair; 
    cellSize: LONG;
    drawDiags: BOOL;
  );
  (*
    Draws a grid of "s[0].ns" by "s[1].ns" cells, with the current
    line style and color.  The grid is intended to be used with
    "MatchInGrid" below.  If "drawDiags" is TRUE, the grid will include
    diagonal lines with the orientation specified by "s[0].rev" and
    "s[1].rev". *)


PROCEDURE MatchBoxes(
    READONLY f: PSPlot.File;
    READONLY m: PZMatch.T;
    READONLY s: SegmentPair; 
    READONLY sRef: SegmentPair; 
    cellSize: LONG; 
    boxSize: REAL;
  );
  (*
    Draws square boxes of side "boxSize" (in mm) around the nodes of
    the the matching "m", with the current line style/color and fill
    color. *)

PROCEDURE MatchPath(
    READONLY f: PSPlot.File;
    READONLY m: PZMatch.T;
    READONLY s: SegmentPair; 
    READONLY sRef: SegmentPair; 
    cellSize: LONG;
  );
  (*
    Draws the steps of the matching "m", as a path in the
    dynamic programming grid, with current line style and color. *)

PROCEDURE MatchDots(
    READONLY f: PSPlot.File;
    READONLY m: PZMatch.T;
    READONLY s: SegmentPair; 
    READONLY sRef: SegmentPair; 
    cellSize: LONG; 
    dotSize: REAL := 0.0;
  );
  (*
    Draws round dots at the nodes of the matching "m", with diameter
    "dotSize" (in mm), with the current line style/color
    and fill color. *)

TYPE
  SegSegPlotProc = PROCEDURE (READONLY t, tRef: SegmentPair);

PROCEDURE TryAllShifts(
    READONLY s: SegmentPair; 
    READONLY sRef: SegmentPair; 
    proc: SegSegPlotProc;
  );
  (*
    Calls "proc(t,tRef)", for some some segment pair "tRef" that is
    equivalent to "sRef" modulo full cycle shift, and every segment
    "t" that is equivalent to "s" and intersects "tRef". *)

END PZPlot.


(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
