MODULE QGet;

IMPORT Rd, Lex;
FROM Lex IMPORT Error;
FROM Thread IMPORT Alerted;
FROM Rd IMPORT Failure, EndOfFile;
FROM Math IMPORT pow;

PROCEDURE Skip(rd: Rd.T; READONLY cs: SET OF CHAR := Blanks) =
  <* FATAL Failure, Alerted *>
  BEGIN Lex.Skip(rd, cs) END Skip;
  
PROCEDURE Match(rd: Rd.T; t: TEXT) =
  <* FATAL Failure, Error, Alerted *>
  BEGIN Lex.Match(rd, t) END Match;
          
PROCEDURE EOL(rd: Rd.T) =
  <* FATAL Failure, Error, Alerted *>
  BEGIN Lex.Skip(rd, Spaces); Lex.Match(rd, "\n") END EOL;
  
PROCEDURE Colon(rd: Rd.T) =
  <* FATAL Failure, Error, Alerted *>
  BEGIN 
    Lex.Skip(rd, Spaces); 
    Lex.Match(rd, ":");
    TRY
      WITH c = Rd.GetChar(rd) DO IF NOT c IN Spaces THEN Rd.UnGetChar(rd) END END
    EXCEPT 
      Rd.EndOfFile => (* OK*) 
    END;
  END Colon;
  
PROCEDURE GetNonBlank(rd: Rd.T): CHAR =
  (* 
    Skip any spaces, then gets next character. Bombs if 
    it is EOF or a non-space blank (TAB, NL, etc.).
  *)
  VAR c: CHAR;
  <* FATAL Failure, Error, Alerted *>
  BEGIN
    Lex.Skip(rd, Spaces); 
    TRY c := Rd.GetChar(rd) EXCEPT EndOfFile => RETURN c END;
    IF c IN Lex.Blanks THEN RAISE Error END;
    RETURN c
  END GetNonBlank;

PROCEDURE Char(rd: Rd.T): CHAR =
  BEGIN 
    RETURN GetNonBlank(rd)
  END Char;

PROCEDURE Bool(rd: Rd.T): BOOLEAN =
  <* FATAL Error *>
  BEGIN 
    WITH c = GetNonBlank(rd) DO
      IF c = 'T' THEN RETURN TRUE
      ELSIF c = 'F' THEN RETURN FALSE
      ELSE RAISE Error
      END
    END
  END Bool;
  
PROCEDURE Int(rd: Rd.T): INTEGER =
  BEGIN
    RETURN DoParseInt(rd, GetNonBlank(rd));
  END Int;

PROCEDURE DoParseInt(rd: Rd.T; c: CHAR): INTEGER =
  (* Parses a signed interger from "c & rd"; leading spaces NOT allowed. *)
  <* FATAL Failure, Error, Alerted *>
  VAR n: INTEGER := 0;
  BEGIN
    IF c = '-' THEN 
      TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile => RAISE Error END;
      IF c < '0' OR c > '9' THEN RAISE Error END;
      REPEAT
        n := 10 * n - (ORD(c) - ORD('0'));
        TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile => RETURN n END;
      UNTIL c < '0' OR c > '9';
      Rd.UnGetChar(rd);
      RETURN n
    ELSE
      IF c = '+' THEN 
        TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile => RAISE Error END
      END;
      IF c < '0' OR c > '9' THEN RAISE Error END;
      REPEAT
        n := 10 * n + (ORD(c) - ORD('0'));
        TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile => RETURN n END;
      UNTIL c < '0' OR c > '9';
      Rd.UnGetChar(rd);
      RETURN n
    END
  END DoParseInt;

PROCEDURE Real(rd: Rd.T): REAL =
  BEGIN 
    RETURN FLOAT(LongReal(rd), REAL)
  END Real;
  
PROCEDURE LongReal(rd: Rd.T): LONGREAL =
  <* FATAL Failure, Error, Alerted *>
  VAR c: CHAR;
      n: LONGREAL := 0.0d0;
      scale: LONGREAL := 1.0d0;
  BEGIN
    c := GetNonBlank(rd);
    
    (* Parse the mantissa as "n / scale". *)
    VAR neg: BOOLEAN;
        empty: BOOLEAN := TRUE;
    BEGIN
      (* Parse sign: *)
      IF c = '-' THEN
        neg := TRUE;
        TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile => RAISE Error END;
      ELSIF c = '+' THEN 
        neg := FALSE;
        TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile => RAISE Error END;
      ELSE
        neg := FALSE
      END;

      (* Parse integer part: *)
      WHILE c >= '0' AND c <= '9' DO
        n := 10.0d0 * n + FLOAT(ORD(c) - ORD('0'), LONGREAL);
        empty := FALSE;
        TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile => 
          IF empty THEN RAISE Error END;
          IF neg THEN RETURN -n ELSE  RETURN n END
        END;
      END;
      IF c = '.' THEN
        (* Parse the fractional part: *)
        TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile => 
          IF empty THEN RAISE Error END;
          IF neg THEN RETURN -n ELSE  RETURN n END
        END;
        WHILE c >= '0' AND c <= '9' DO
          n := n * 10.0d0 +  FLOAT(ORD(c) - ORD('0'), LONGREAL);
          scale := scale * 10.0d0;
          empty := FALSE;
          TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile =>
            IF empty THEN RAISE Error END;
            IF neg THEN RETURN -n/scale ELSE RETURN n/scale END
          END;
        END;
      END;
      IF empty THEN RAISE Error END;
      IF neg THEN n := -n END;
    END;
    
    IF c # 'd' AND c # 'D' AND c # 'e' AND c # 'E' THEN
      Rd.UnGetChar(rd);
      RETURN n / scale
    ELSE
      (* Parse exponent, add to current "exp": *)
      TRY c := Rd.GetChar(rd) EXCEPT Rd.EndOfFile => RETURN n END;
      VAR exp := DoParseInt(rd, c);
      BEGIN
        IF exp > 10 THEN
          RETURN n * (pow(10.0d0, FLOAT(exp, LONGREAL)) / scale)
        ELSIF exp > 0 THEN
          WHILE exp > 0 DO 
            IF scale > 1.0d0 THEN scale := scale/10.0d0 ELSE n := n * 10.0d0 END;
            DEC(exp)
          END;
          RETURN n / scale
        ELSIF exp < -10 THEN
          RETURN n / (pow(10.0d0, FLOAT(-exp, LONGREAL)) * scale)
        ELSIF exp < 0 THEN
          WHILE exp < 0 DO 
            scale := scale*10.0d0;
            INC(exp)
          END;
          RETURN n / scale
        ELSE
          RETURN n
        END
      END
    END        
  END LongReal;

BEGIN
END QGet.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
